##  @package seal.core.misc
#   Miscellaneous useful Python extensions.

import datetime, os, pty, unicodedata, sys, imp, threading, traceback, importlib
import signal
from itertools import islice, chain
from importlib import import_module
from io import StringIO
from time import time
from math import inf
import collections


#--  General  ------------------------------------------------------------------

##  Checks whether the object has the given type.

def check_type (x, t):
    xnames = [c.__name__ for c in x.__class__.__mro__]
    if isinstance(t, (tuple, list)):
        types = t
    else:
        types = [t]
    for ty in types:
        if isinstance(ty, str):
            name = ty
        else:
            name = ty.__name__
        if name in xnames:
            return
    raise Exception('Object does not have required type: %s, %s' % (x, t))


#--  Matching  -----------------------------------------------------------------

##  Indicates whether the object matches the description.

def matches (x, descr):
    for (k, v) in descr.items():
        if v is None: continue
        if not hasattr(x, k): return False
        val = getattr(x, k)
        if isinstance(val, collections.Callable): val = val()
        if isinstance(v, list):
            if val not in v: return False
        else:
            if val != v: return False
    return True


##  Returns a module given its name.
#   May raise ModuleNotFoundError

def string_to_module (s):
    if not s:
        raise Exception('Require nonempty name')
    return import_module(s)

##  Takes a fully-qualified name and gets the object.

def string_to_object (s):
    j = s.rfind('.')
    if j < 0:
        raise Exception('Require fully qualified name')
    m = string_to_module(s[:j])
    return m.__dict__[s[j+1:]]


#--  Object  -------------------------------------------------------------------

class Object (object):

    def __init__ (self, *args, **kwargs):
        d = dict(kwargs)
        object.__setattr__(self, '_dict', d)
        if d:
            object.__setattr__(self, '_listlen', None)
            for (i, arg) in enumerate(args):
                d[i] = arg
        else:
            object.__setattr__(self, '_listlen', len(args))
            self.extend(args)

    def _setlistlen (self, n):
        object.__setattr__(self, '_listlen', n)

    def __getitem__ (self, key):
        return self._dict[key]

    def __getattr__ (self, name):
        return self._dict[name]
            
    def __setitem__ (self, name, value):
        self._dict[name] = value
        n = self._listlen
        if not (isinstance(name, int) and n is not None and 0 <= name < n):
            self._setlistlen(None)

    def __setattr__ (self, name, value):
        self.__setitem__(name, value)

    def append (self, value):
        if self._listlen is None:
            raise Exception('Not a list-like Object')
        else:
            n = self._listlen
            self._setlistlen(n + 1)
            self._dict[n] = value

    def extend (self, values):
        for val in values:
            self.append(val)

    def __iter__ (self):
        if self._listlen is None:
            for k in self._dict.keys():
                yield k
        else:
            for i in range(self._listlen):
                yield self._dict[i]

    def __len__ (self): return len(self._dict)
    def items (self): return self._dict.items()
    def keys (self): return self._dict.keys()
    def values (self): return self._dict.values()

    def __repr__ (self):
        if self._listlen is None:
            return repr(self._dict)
        else:
            return repr(list(self))


#--  Reflexion  ----------------------------------------------------------------

def _docstring_lines (x):
    doc = x.__doc__
    if doc:
        if doc.startswith('\n'):
            i = 1
            while i < len(doc) and doc[i].isspace() and doc[i] not in '\r\n':
                i += 1
            prefix = doc[1:i]
            doc = doc[i:]
        else:
            prefix = ''
        n = len(prefix)
        for line in doc.split('\n'):
            if line.startswith(prefix):
                line = line[n:]
            yield line


class FunctionInfo (object):

    def __init__ (self, fnc, ismethod=False):
        dflts = fnc.__defaults__ or []
        nkws = len(dflts)
        varnames = fnc.__code__.co_varnames
        nselfargs = 1 if ismethod else 0
        nargs = fnc.__code__.co_argcount - (nselfargs + nkws)
        i = nselfargs
        j = 1 + nargs
        k = j + nkws   # after kws are local variables
        args = varnames[i:j]
        kws = varnames[j:k]
        doc = list(_docstring_lines(fnc)) if fnc.__doc__ else []

        self.function = fnc
        self.args = args
        self.kwargs = list((kws[i], dflts[i]) for i in range(len(kws)))
        self.doc = doc


def MethodInfo (method):
    return FunctionInfo(method, ismethod=True)


#--  ListProxy, MapProxy  ------------------------------------------------------

class ListProxy (object):

    def __iter__ (self):
        return self.__list__().__iter__()
        
    def __contains__ (self, k):
        return self.__list__().__contains__(k)

    def __getitem__ (self, k):
        return self.__list__().__getitem__(k)

    def __len__ (self):
        return self.__list__().__len__()

    def __repr__ (self):
        return self.__list__().__repr__()


class MapProxy (object):

    def __iter__ (self):
        return iter(self.__map__())
        
    def __len__ (self):
        return len(self.__map__())

    def __contains__ (self, k):
        return k in self.__map__()

    def __getitem__ (self, k):
        return self.__map__()[k]

    def get (self, k, dflt=None):
        return self.__map__().get(k, dflt)

    def keys (self):
        return self.__map__().keys()

    def values (self):
        return self.__map__().values()
    
    def items (self):
        return self.__map__().items()

    def __repr__ (self):
        return self.__map__().__repr__()


class LazyList (ListProxy):

    # The iterf should be repeatedly callable, returning an iteration
    # over the underlying data structure each time

    def __init__ (self, iterf):
        self._expanded = None
        self._iterf = iterf

    def __list__ (self):
        if self._expanded is None:
            self._expanded = list(self._iterf())
        return self._expanded

    def __iter__ (self):
        if self._expanded is None:
            return self._iterf()
        else:
            return iter(self._expanded)

    def __repr__ (self):
        if self._expanded is None:
            return '[...]'
        else:
            return repr(self._expanded)




from math import inf
from itertools import islice


#--  Lists  --------------------------------------------------------------------

##  Converts x to a list.

def as_list (x):
    if x == None: return []
    elif isinstance(x, list): return x
    elif isinstance(x, (tuple, set, frozenset, range, dict)): return list(x)
    elif hasattr(x, '__next__'): return list(x)
    else: return [x]

##  Returns an iterable.

def repeatable (x):
    if x == None: return []
    elif hasattr(x, '__next__'): return iter(x)
    elif hasattr(x, '__iter__'): return x
    else: raise Exception('Does not appear to be iterable: %s' % repr(x))

##  Concatenates multiple lists.

def concat (lists):
    out = []
    for list in lists:
        out.extend(list)
    return out

##  Eliminates duplicates from a list.  Otherwise preserves order.

def unique (elts):
    out = []
    for elt in elts:
        if elt not in out:
            out.append(elt)
    return out

##  Cross product.

def cross_product (lists):
    if len(lists) == 1:
        return [(x,) for x in lists[0]]
    else:
        return [(x,) + rest
                for x in lists[0]
                for rest in cross_product(lists[1:])]


#--  Edit distance  ------------------------------------------------------------

##  Works with arbitrary sequences; requires len and [].
#   The cost function should accept pairs of elements (x,y).
#       (x,None) = deletion of x
#       (None,y) = insertion of y
#       (x,y) otherwise = substitution of y for x.  Possibly x=y.
#   A "cell" represents a pair of positions: i in the input, j in the output.
#   Both i and j range from 0 to the end of their sequence, inclusive.
#   i + j is the "norm" of the cell.  To fill in a cell, the cells we need
#   all have smaller norms, so we guarantee they have already been filled by
#   visiting by increasing norm.

class EDNode (object):

    def __init__ (self, i, j):
        self.i = i
        self.j = j
        self.cost = None
        self.prev = None

    def __repr__ (self):
        return '<EDNode %d %d>' % (self.i, self.j)


def simple_cost (x, y):
    if x != y: return 1
    else: return 0


class EditDistance (object):

    def __init__ (self, cost=simple_cost):
        self.cost_function = cost

    def table (self, inseq, outseq):
        cost = self.cost_function
        m = len(inseq)
        n = len(outseq)
        nc = n + 1
        maxnorm = m + n

        nodes = [EDNode(i,j) for i in range(m+1) for j in range(n+1)]
        nodes[0].cost = 0
    
        for norm in range(1, maxnorm + 1):  # include the case norm = maxnorm
            imin = 0 if norm < n else norm - n
            imax = m if m < norm else norm
            for i in range(imin, imax + 1):  # include the case i = imax
                j = norm - i

                node = nodes[i*nc + j]
                node.cost = inf
                node.prev = None
    
                # substitution
                if i > 0 and j > 0:
                    prev = nodes[(i-1)*nc + (j-1)]
                    c = prev.cost + cost(inseq[i-1], outseq[j-1])
                    if c < node.cost:
                        node.cost = c
                        node.prev = prev
    
                # deletion
                if i > 0:
                    prev = nodes[(i-1)*nc + j]
                    c = prev.cost + cost(inseq[i-1], None)
                    if c < node.cost:
                        node.cost = c
                        node.prev = prev
    
                # insertion
                if j > 0:
                    prev = nodes[i*nc + (j-1)]
                    c = prev.cost + cost(None, outseq[j-1])
                    if c < node.cost:
                        node.cost = c
                        node.prev = prev
    
        return nodes

    def __call__ (self, inseq, outseq):
        nodes = self.table(inseq, outseq)
        return nodes[-1].cost


#--  Sorted lists  -------------------------------------------------------------

##  Eliminates duplicates, assuming that the list is sorted.

def uniq (sortedlst):
    out = []
    for elt in sortedlst:
        if len(out) == 0 or elt != out[-1]:
            out.append(elt)
    return out

##  Intersects two sorted lists.  Unpredictable results if the lists are not
#   sorted.

def intersect (list1, list2):
    out = []
    i1 = 0
    i2 = 0
    n1 = len(list1)
    n2 = len(list2)
    while i1 < n1 and i2 < n2:
        if list1[i1] < list2[i2]:
            i1 += 1
        elif list2[i2] < list1[i1]:
            i2 += 1
        else:
            out.append(list1[i1])
            i1 += 1
            i2 += 1
    return out

##  Takes the union of two sorted lists.  Unpredictable results if the lists
#   are not sorted.

def union (list1, list2):
    out = []
    i1 = 0
    i2 = 0
    n1 = len(list1)
    n2 = len(list2)
    while i1 < n1 and i2 < n2:
        if list1[i1] < list2[i2]:
            out.append(list1[i1])
            i1 += 1
        elif list2[i2] < list1[i1]:
            out.append(list2[i2])
            i2 += 1
        else:
            out.append(list1[i1])
            i1 += 1
            i2 += 1
    # only one of these loops will actually be used
    while i1 < n1:
        out.append(list1[i1])
        i1 += 1
    while i2 < n2:
        out.append(list2[i2])
        i2 += 1
    return out

##  Returns the set difference of two sorted lists.  Unpredictable results if
#   the lists are not sorted.

def difference (list1, list2):
    out = []
    i1 = 0
    i2 = 0
    n1 = len(list1)
    n2 = len(list2)
    while i1 < n1 and i2 < n2:
        if list1[i1] < list2[i2]:
            out.append(list1[i1])
            i1 += 1
        elif list2[i2] < list1[i1]:
            i2 += 1
        else:
            i1 += 1
            i2 += 1
    while i1 < n1:
        out.append(list1[i1])
        i1 += 1
    return out


#--  Queue  --------------------------------------------------------------------

##  A queue.  It uses a circular buffer.

class Queue (object):

    ##  Constructor.

    def __init__ (self, maxwaste=10):
        self.__buffer = []
        self.__head = 0
        self.__tail = 0
        self.__maxwaste = maxwaste

    ##  The number of elements.

    def __len__ (self):
        return self.__tail - self.__head

    ##  Whether the queue contains any elements.

    def __bool__ (self):
        return self.__tail > self.__head

    def __buffer_index (self, i):
        if isinstance(i, int):
            return self.__head + i
        elif isinstance(i, slice):
            if i.start: start = self.__head + i.start
            else: start = self.__head
            if i.stop: stop = self.__head + i.stop
            else: stop = self.__tail
            return slice(start, stop, i.step)
        else:
            raise IndexError('Not a valid index: %s' % repr(i))

    ##  Fetch the item at the given queue position.

    def __getitem__ (self, i):
        return self.__buffer.__getitem__(self.__buffer_index(i))

    ##  Set the item at the given queue position.

    def __setitem__ (self, i, v):
        self.__buffer.__setitem__(self.__buffer_index(i), v)

    ##  Add an element to the end of the queue.

    def write (self, elt):
        if self.__tail == len(self.__buffer):
            self.__buffer.append(elt)
        else:
            self.__buffer[self.__tail] = elt
        self.__tail += 1

    ##  Return the item at the head of the queue and advance the head.

    def read (self):
        if self.__head >= self.__tail:
            raise IndexError('Empty queue')
        elt = self.__buffer[self.__head]
        self.__head += 1
        if self.__head == self.__tail or \
                self.__maxwaste is not None and self.__head > self.__maxwaste:
            n = self.__tail - self.__head
            for i in range(n):
                self.__buffer[i] = self.__buffer[self.__head + i]
            self.__head = 0
            self.__tail = n
        return elt


#--  Iterables  ----------------------------------------------------------------

##  Returns the n-th item of iterable g, counting from 0, and counting from the
#   current position of the iterator's "read head."  For example:
#
#       >>> import itertools
#       >>> c = itertools.count(0)
#       >>> nth(c, 3)
#       3
#       >>> nth(c, 3)
#       7
#
#   The iterator I{c} generates the natural numbers, beginning with 0.  The
#   first call to C{nth} returns the fourth item, which is the number 3.
#   The second call begins where the first left off, and returns the
#   fourth item, which is the number 7.
#
#   Note that one can achieve the same functionality this way:
#
#       >>> itertools.islice(c, 3, 4).next()
#
#   One use of nth is to jump to problematic cases in a large
#   iteration.  An idiom for finding such cases in the first place is the
#   following:
#    
#       for i, x in enumerate(myiteration):
#           if isproblematic(x):
#               return i

def nth (iter, n):
    try:
        count = 0
        while count < n:
            next(iter)
            count += 1
        return next(iter)
    except StopIteration:
        return None

##  Returns a list containing the first n elements from the iteration.

def head (iter, n=5):
    return list(islice(iter, n))

##  Returns a list containing the last n elements from the iteration.

def tail (iter, n=5):
    out = []
    ptr = 0
    for elt in iter:
        if len(out) < n:
            out.append(elt)
        else:
            out[ptr] = elt
        ptr += 1
        if ptr >= n: ptr = 0

    if len(out) < n or ptr == 0:
        return out
    else:
        return out[ptr:] + out[:ptr]


##  A pager.
class _Pager (object):

    ##  Constructor.
    def __init__ (self, pagesize=40):
        ##  The page size, in lines.
        self.pagesize = pagesize

    ##  Call it.
    def __call__ (self, iter):
        for i,x in enumerate(iter):
            if i > 0 and i % self.pagesize == 0:
                if input() == 'q': break
            print(x)

##  Prints a "page" at a time and waits for space bar or 'q'.
more = _Pager()

##  The product of a list of numbers.

def product (nums):
    prod = 1
    for num in nums:
        prod *= num
    return prod

##  Counts up how many items are contained in
#   (or remain in) the given iterable.  It calls C{next()} repeatedly
#   until the iteration is "used up."  If the iterable is infinite, it
#   never returns.  It is defined as: python::
#
#       sum(1 for x in g)
#

def count (iter):
    return sum(1 for x in iter)


##  Creates a map whose keys are items in the given iterable, and whose value for
#   a given item is the frequency of occurrence of that item in the iteration.
#   All items are consumed.

def counts (iterable):
    dict = {}
    for item in iterable:
        if item in dict:
            dict[item] += 1
        else:
            dict[item] = 1
    return dict

def dups (items):
    counts = {}
    for item in items:
        if item in counts:
            if counts[item] == 1:
                yield item
            counts[item] += 1
        else:
            counts[item] = 1




import datetime, unicodedata
from io import StringIO
from time import time as now


#--  Strings  ------------------------------------------------------------------

##  Describe each character in a unicode string.

def unidescribe (s):
    for (i,c) in enumerate(s):
        print(i, hex(ord(c)), unicodedata.name(c))


##  A <b>word</b> consists only of alphanumerics and underscore, and
#   is not the empty string.

def isword (s):
    return len(s) > 0 and all(c.isalnum() or c == '_' for c in s)


##  Break the string into its lines.  The lines do not include carriage return
#   and newline.

def lines (s):
    if isinstance(s, (bytes, bytearray)):
        cr = 13
        nl = 10
    else:
        cr = '\r'
        nl = '\n'
    i = 0
    while True:
        k = s.find(nl, i)
        if k < 0: break
        j = k
        if j > i and s[j-1] == cr:
            j -= 1
        yield s[i:j]
        i = k + 1


#--  as_ascii  -------------------------

def _objectionable_codepoint (x):
    return (x < 32 or x == 123 or x == 125 or x >= 127)

def _is_unobjectionable (s):
    if isinstance(s, str):
        for c in s:
            x = ord(c)
            if _objectionable_codepoint(x): return False
        return True
    else:
        return False

_my_names = {7: 'bel',
             8: 'bs',
             9: 'tab',
             10: 'nl',
             11: 'vt',
             12: 'ff',
             13: 'ret',
             27: 'esc',
             123: 'lb',
             125: 'rb',
             127: 'del',
             8211: 'end', # 2013
             8212: 'emd', # 2014
             8216: 'lsq', # 2018
             8217: 'rsq', # 2019
             8220: 'ldq', # 201c
             8221: 'rdq'} # 201d

_my_names_inv = None

_substitutions = {9: ' ',    # tab
                  11: '\n',  # vert tab
                  12: '\n',  # form feed
                  8211: '-', # en-dash
                  8212: '-', # em-dash
                  8216: "'", # left single quote
                  8217: "'", # right single quote
                  8220: '"', # left double quote
                  8221: '"'} # right double quote

def _write_name (c, x, out):
    try:
        out.write(unicodedata.name(c))
    except Exception:
        out.write(hex(x)[2:])

def _write_hex (c, x, out):
    s = hex(x)[2:]
    if len(s) == 0:
        raise Exception('Empty hex string')
    elif len(s) == 1:
        out.write('0')
        out.write(s)
    elif len(s) == 2:
        out.write(s)
    else:
        for i in range(4-len(s)):
            out.write('0')
        out.write(s)

##  Convert a string to ASCII.

def as_ascii (s, use='hex'):
    if not isinstance(s, str):
        return as_ascii(str(s), use)
    elif _is_unobjectionable(s):
        return s
    if use in (None, 'alts'):
        with StringIO() as out:
            for c in s:
                x = ord(c)
                if x == 10 or 32 <= x <= 126:
                    out.write(c)
                elif use == 'alts' and x in _substitutions:
                    out.write(_substitutions[x])
            return out.getvalue()
    else:
        if use == 'names': f = _write_name
        elif use == 'hex': f = _write_hex
        else: raise Exception('Bad value for use: %s' % use)
        with StringIO() as out:
            for c in s:
                x = ord(c)
                if _objectionable_codepoint(x):
                    out.write('{')
                    if x in _my_names:
                        out.write(_my_names[x])
                    else:
                        f(c, x, out)
                    out.write('}')
                else:
                    out.write(c)
            return out.getvalue()

##  Undo as_ascii().

def from_ascii (s):
    global _my_names_inv
    if '{' in s:
        out = StringIO()
        i = 0
        while i < len(s):
            c = s[i]
            if c == '{':
                i += 1
                j = s.find('}', i)
                if j < 0:
                    raise Exception("'{' without matching '}'")
                name = s[i:j]
                if name.isdigit():
                    c = chr(int(name, 16))
                else:
                    if _my_names_inv is None:
                        _my_names_inv = {}
                        for (k,v) in _my_names.items():
                            _my_names_inv[v] = chr(k)
                    c = _my_names_inv[name]
                out.write(c)
                i = j+1
            else:
                out.write(c)
                i += 1
        s = out.getvalue()
        out.close()
        return s
    else:
        return s


##  Convert a string to printable ASCII characters.  Here, a "printable"
#   character is a character in the range U+0020 (space) to U+007e inclusive,
#   or newline (U+000a).  That is, space and newline are considered printable,
#   all other whitespace and control characters are not; nor is DEL (U+007f).
#   Use substitutions for "smart quotes" and related characters.  Replace tab
#   with space; replace vertical tab, and formfeed with newline.  Delete all
#   other characters that are not printable ASCII.  Returns an iteration over
#   characters.

def ascii_chars (s):
    for c in s:
        n = ord(c)
        if n in _substitutions:
            yield _substitutions[n]
        elif n >= 0x20 and n <= 0x7e:
            yield c
        elif n == 0x0a:
            yield c

##  Convert a single character to a (possibly escaped) form that is safe for
#   inclusion between double quotes.  Double quotes and backslashes are escaped
#   with backslashes, and newline is replaced with backslash-en.  Returns an
#   iteration over strings.

def quotable_chars (s):
    for c in s:
        if c == '"':
            yield '\\"'
        elif c == '\\':
            yield '\\\\'
        elif c == '\n':
            yield '\\n'
        else:
            yield c

##  Like repr(), but always returns a double-quoted string.  May be called as:
#   quote(ascii_chars(s)).

def quoted (s):
    with StringIO() as f:
        f.write('"')
        for c in quotable_chars(s):
            f.write(c)
        f.write('"')
        return f.getvalue()


#--  deaccent  -------------------------

_deaccent_map = [
    None, None, None, None, None, None, None, None, None, '\t', '\n', None, None, '\r', None, None,
    None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
    ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/',
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?',
    '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_',
    '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', None,
    None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
    None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None,
    None, '!', 'c', 'L', 'O', 'Y', None, None, None, 'c', 'a', '<<', '-', None, 'R', None,
    None, '+/-', '2', '3', None, 'm', None, '.', None, '1', 'o', '>>', '1/4', '1/2', '3/4', '?',
    'A', 'A', 'A', 'A', 'A', 'A', 'AE', 'C', 'E', 'E', 'E', 'E', 'I', 'I', 'I', 'I',
    'Dh', 'N', 'O', 'O', 'O', 'O', 'O', '*', 'O', 'U', 'U', 'U', 'U', 'Y', 'Th', 'ss',
    'a', 'a', 'a', 'a', 'a', 'a', 'ae', 'c', 'e', 'e', 'e', 'e', 'i', 'i', 'i', 'i',
    'dh', 'n', 'o', 'o', 'o', 'o', 'o', '/', 'o', 'u', 'u', 'u', 'u', 'y', 'th', 'y'
    ]

##  Map accented characters to their unaccented form.

def deaccent (s):
    asc = []
    for u in s:
        i = ord(u)
        if i >= 0 and i < len(_deaccent_map):
            a = _deaccent_map[i]
            if a:
                asc.append(a)
    return ''.join(asc)


##  Interpret an ASCII string as UTF8.

def utf8 (s, fn=None):
    if fn is None:
        print(' '.join(hex(c)[2:] for c in s.encode('utf8')))
    else:
        with open(fn, 'w', encoding='utf8') as f:
            print(s, file=f)

##  Convert to a boolean value.

def as_boolean (s):
    if s == 'False': return False
    elif s == 'True': return True
    else: raise Exception('Not a boolean: ' + str(s))

##  Convert the string to ASCII and trim it to fit in a field with
#   fixed width w.

def trim (w, s=None):
    if s is None:
        s = w
        w = 0
    s = as_ascii(s, use='hex')
    if w > 0 and len(s) > w: return s[:w]
    else: return s


#--  Date-Time string, Size string  --------------------------------------------

##  Print out a timestamp readably.

def dtstr (timestamp):
    return datetime.datetime.fromtimestamp(timestamp).isoformat().replace('T',' ')

##  Print out a file size readably.

def sizestr (nbytes):
    if nbytes >= 1000000000000000:
        return '%5.3f PB' % (nbytes/1000000000000000)
    elif nbytes >= 1000000000000:
        return '%5.3f TB' % (nbytes/1000000000000)
    elif nbytes >= 1000000000:
        return '%5.3f GB' % (nbytes/1000000000)
    elif nbytes >= 1000000:
        return '%5.3f MB' % (nbytes/1000000)
    elif nbytes >= 1000:
        return '%5.3f KB' % (nbytes/1000)
    else:
        return '%d B' % nbytes


##  Time difference, in human-readable form.

def elapsed_time_str (start, end):
    sign = ''
    if end < start:
        sign = '-'
        (start, end) = (end, start)
    return sign + timestr(end - start)

##  Readable string showing an amount of time.

def timestr (nsec):
    nhr = 0
    nmin = 0
    if nsec < 1:
        return '%.4f ms' % (nsec * 1000)
    if nsec >= 60:
        nmin = int(nsec / 60)
        nsec -= nmin * 60
        if nmin >= 60:
            nhr = int(nmin / 60)
            hstr = int(nhr)
            nmin -= nhr * 60
    intsec = int(nsec)
    fraction = nsec - intsec
    return '%d:%02d:%02d%s' % (nhr, nmin, intsec, ('%.4f' % fraction)[1:])


class Time (object):

    def __init__ (self):
        self.t0 = None

    def __enter__ (self):
        self.t0 = now()

    def __exit__ (self, t, v, tb):
        t1 = now()
        print('Time elapsed (h:mm:ss):', timestr(t1 - self.t0))


#--  Expand environment variables  ---------------------------------------------

def expand_envvars (s):
    out = []
    i = 0
    while True:
        j = s.find('$', i)
        if j < 0:
            if i < len(s):
                out.append(s[i:])
            break
        # s[j] is a $
        if j > i: out.append(s[i:j])
        i = j + 1
        if s[i] == '$':
            out.append('$')
            i += 1
        elif s[i] in '({':
            if s[i] == '(': end = ')'
            else: end = '}'
            i += 1
            j = s.find(end, i)
            name = s[i:j].strip()
            if not name: raise Exception('Empty env var name')
            v = os.environ.get(name)
            if v: out.append(v)
            i = j + 1
        elif s[i].isalpha() or s[i] == '_':
            j = i + 1
            while j < len(s) and (s[j].isalnum() or s[j] == '_'):
                j += 1
            name = s[i:j].strip()
            if not name: raise Exception('Empty env var name')
            v = os.environ.get(name)
            if v: out.append(v)
            i = j
        else:
            raise Exception('Illegal character after $')
    return ''.join(out)


#--  Reflection  ---------------------------------------------------------------

##  Returns a module given its name.
#   May raise ModuleNotFoundError

def string_to_module (s):
    if not s:
        raise Exception('Require nonempty name')
    return import_module(s)

##  Takes a fully-qualified name and gets the object.

def string_to_object (s):
    j = s.rfind('.')
    if j < 0:
        raise Exception('Require fully qualified name')
    m = string_to_module(s[:j])
    return m.__dict__[s[j+1:]]




import sys, subprocess
from time import time
from selkie.string import elapsed_time_str


#--  file_size  ----------------------------------------------------------------

##  Returns the file size in bytes.

def file_size (fn):
    return os.stat(fn).st_size


#--  System calls  -------------------------------------------------------------

##  Make a synchronous call, raising an exception on failure.
#   If the call is unsuccessful, an exception is raised.
#   There is no return value.

# def call (program, *args):
#     value = os.spawnvp(os.P_WAIT, program, [program] + list(args))
#     if value != 0:
#         raise Exception("Program failed: value " + str(value) + ": " + program + " " + ' '.join(args))

##  Makes an asynchronous call, returning 0 on success and an error code on failure.

# def launch (program, *args):
#     return os.spawnvp(os.P_NOWAIT, program, [program] + list(args))


def system (*args, silent=False):
    return True if subprocess.run(args, capture_output=silent).returncode == 0 else False

# return value has members returncode, stdout, stderr

def backtick (*args):
    return subprocess.run(args, check=True, capture_output=True, text=True)


#--  Shell  --------------------------------------------------------------------

##  Prints output all at once, not incrementally.

def shell (cmd):
    print(subprocess.getoutput(cmd))

##  Unless silent=True, prints incrementally.
#   @arg value - '', 'v', 'vo', 'o'.
#   Unless 'v', signals an error for nonzero status.
#    - Returns (value, output) if 'vo'.
#    - Returns value if 'v'.
#    - Returns output if 'o'.

def run_command (com, *args, value='', silent=False, color=None):
    if color is None:
        return _run_command1(com, args, value, silent)
    else:
        with _stdout_color(color):
            return _run_command1(com, args, value, silent)

def _run_command1 (com, args, value, silent):
    if 'o' in value: output = bytearray()
    (pid, fd) = pty.fork()
    if pid == 0:
        os.execlp(com, com, *args)
    else:
        if not silent: os.write(1, b'\033[36m')
        while True:
            b = _read(fd, 64)
            if not b: break
            if not silent: os.write(1, strip_escapes(b))
            if 'o' in value: output.extend(b)
        if not silent: os.write(1, b'\033[39m')
        status = os.waitpid(pid, 0)[1]
        if not ('v' in value or status == 0):
            raise Exception('Executable error: %s' % status)
        if 'o' in value:
            output = output.decode('raw_unicode_escape')
        if value == 'vo': return (status, output)
        elif value == 'v': return status
        elif value == 'o': return output

def _read (fd, bufsize):
    b = b''
    try:
        b = os.read(fd, bufsize)
    except OSError:
        pass
    return b

##  Stdout color.
class _stdout_color (object):
    ##  Constructor.
    def __init__ (self, color):
        colors = {'red': b'\033[31m',
                  'green': b'\033[32m',
                  'yellow': b'\033[33m',
                  'blue': b'\033[34m',
                  'magenta': b'\033[35m',
                  'cyan': b'\033[36m'}
        ##  The escape code for the given color.
        self.color = colors[color]
    ##  Enter.
    def __enter__ (self):
        os.write(1, self.color)
    ##  Exit.
    def __exit__ (self, t, v, tb):
        os.write(1, b'\033[39m')


def _is_digit_or_semicolon (x):
    return (48 <= x <= 57 or x == 59)

def _is_letter (x):
    return (65 <= x <= 90 or 97 <= x <= 122)

# returns end point if looking at xterm escape sequence, -1 otherwise

def _escape_seq (b, i):
    if not (i < len(b) and b[i] == 27): return -1
    i += 1
    if not (i < len(b) and b[i] == 91): return i
    i += 1
    while i < len(b) and _is_digit_or_semicolon(b[i]): i += 1
    if i < len(b) and _is_letter(b[i]): i += 1
    return i

##  Strip xterm escape sequences, e.g. color changes.

def strip_escapes (b):
    out = None
    i = k = 0
    while k < len(b):
        j = _escape_seq(b,k)
        if j < 0:
            k += 1
        else:
            if out is None: out = bytearray(b[:k])
            else: out.extend(b[i:k])
            i = k = j
    if i < len(b) and out is not None:
        out.extend(b[i:])
    if out is None: return b
    else: return out


#--  More  ---------------------------------------------------------------------

##  A pager.
class _Pager (object):

    ##  Constructor.
    def __init__ (self, pagesize=40):
        ##  The page size, in lines.
        self.pagesize = pagesize

    ##  Call it.
    def __call__ (self, iter):
        for i,x in enumerate(iter):
            if i > 0 and i % self.pagesize == 0:
                if input() == 'q': break
            print(x)

##  Prints a "page" at a time and waits for space bar or 'q'.
more = _Pager()


#--  Command line processing, v1  ----------------------------------------------

##  Deprecated.  Use shift instead.

class CommandLine (object):

    ##  Constructor.

    def __init__ (self, usage, nargs=None):
        self.__usage = usage
        ##  Arg vector.
        self.argv = sys.argv
        ##  Current position.
        self.i = 1
        ##  Command.
        self.arg0 = None
        ##  Number of arguments.
        self.nargs = nargs

    ##  Print usage and exit.
    def usage (self):
        sys.stderr.write('Usage: %s\n' % self.__usage)
        sys.exit(1)

    ##  Whether the current argument is a flag.
    def has_option (self):
        if len(self.argv) > self.i and self.argv[self.i][0] == '-':
            return True
        else:
            return False

    ##  Get the value for opt.
    def option (self, opt):
        if len(self.argv) <= self.i or self.argv[self.i][0] != '-':
            raise Exception('No option')
        key = self.argv[self.i]
        self.i += 1
        return key

    ##  Check whether the number of arguments is correct.
    def check_nargs (self):
        if self.arg0 is None:
            self.arg0 = self.i
            if self.nargs is None:
                self.nargs = len(self.argv) - self.arg0
            elif len(self.argv) - self.arg0 > self.nargs:
                self.usage()

    ##  Iterator method.
    def __next__ (self):
        self.check_nargs()
        t = self.i - self.arg0
        if t >= self.nargs:
            raise StopIteration
        if t < len(self.argv):
            arg = self.argv[t]
        else:
            arg = ''
        self.i += 1
        return arg

    ##  Get the i-th argument.
    def __getitem__ (self, i):
        self.check_nargs()
        if i < 0 or i >= self.nargs: raise IndexError
        t = self.arg0 + i
        if t < len(self.argv):
            return self.argv[t]
        else:
            return ''

    ##  Get the number of arguments.
    def __len__ (self):
        self.check_nargs()
        return self.nargs

    ##  Returns self.
    def __iter__ (self):
        return self


#--  Command line processing, v2  ----------------------------------------------

def _n_oblig_params (fnc):
    nparams = fnc.__code__.co_argcount
    nopt = len(fnc.__defaults__)
    return nparams - nopt

def _optional_params (fnc):
    nparams = fnc.__code__.co_argcount
    nopt = len(fnc.__defaults__)
    noblig = nparams - nopt
    names = fnc.__code__.co_varnames
    return names[noblig:nparams]

##  Deprecated.  Use shift instead.

def run_main_2 (main):
    if len(sys.argv) > 1 and sys.argv[1] == '-?': _print_usage(main)
    show_stack_trace = False
    kwargs = {}
    noblig = _n_oblig_params(main)
    i = 1 + noblig
    if i > len(sys.argv): _print_usage(main)
    while i < len(sys.argv):
        arg = sys.argv[i]
        i += 1
        if arg == '-!': show_stack_trace = True
        elif arg == '--': break
        else:
            k = arg.find('=')
            if k >= 0:
                key = arg[1:k]
                value = arg[k+1:]
            else:
                key = arg[1:]
                value = '1'
            if key in kwargs:
                print("Flag '-%s' multiply specified" % key, file=sys.stderr)
                sys.exit(1)
            kwargs[key] = value

    args = sys.argv[i:]
    try:
        value = main(*args, **kwargs)
        if value is not None:
            print(value)
        sys.exit(0)
    except Exception as e:
        if show_stack_trace:
            traceback.print_exc()
        else:
            print('ERROR:', e, file=sys.stderr)
        sys.exit(1)


#--  Command line processing, v3  ----------------------------------------------

##  Deprecated.  Use shift instead.

def run_main (main):
    show_stack_trace = False
    kwargs = {}
    i = 1
    while i < len(sys.argv) and sys.argv[i].startswith('-'):
        arg = sys.argv[i]
        i += 1
        if arg == '-?': _print_usage(main)
        elif arg == '-!': show_stack_trace = True
        elif arg == '--': break
        else:
            k = arg.find('=')
            if k >= 0:
                key = arg[1:k]
                value = arg[k+1:]
            else:
                key = arg[1:]
                value = '1'
            if key in kwargs:
                print("Flag '-%s' multiply specified" % key, file=sys.stderr)
                sys.exit(1)
            kwargs[key] = value

    args = sys.argv[i:]
    try:
        value = main(*args, **kwargs)
        if value is not None:
            print(value)
        sys.exit(0)
    except Exception as e:
        if show_stack_trace:
            traceback.print_exc()
        else:
            print('ERROR:', e, file=sys.stderr)
        sys.exit(1)

def _print_usage (f):
    progname = sys.argv[0]
    nargs = f.__code__.co_argcount
    defaults = f.__defaults__
    nopt = len(defaults)
    noblig = nargs - nopt
    allow_xs_positional = (f.__code__.co_flags & 0x04)
    allow_xs_keywords = (f.__code__.co_flags & 0x08)
    varnames = f.__code__.co_varnames
    oblig = varnames[:noblig]
    opt = varnames[noblig:nargs]
    words = [progname] + list(oblig)
    if opt: words.append('[%s]' % ' '.join(opt))
    print('USAGE:   ', ' '.join(words))
    if defaults:
        print('Defaults:')
        for i in range(nopt):
            print('    %-10s %s' % (opt[i], defaults[i]))
    print('Flags:')
    print('    Any arg can be provided as a flag in form -arg=value')
    print('    -arg  is the same as -arg=1')
    print('    -?    prints this usage message')
    print('    -!    causes a stack trace to be printed in case of error')
    print('    --    terminates flags; what follows are positional arguments')
    if f.__doc__:
        print()
        print(f.__doc__)
    sys.exit(1)


#--  Command line processing, v4  ----------------------------------------------

def _print_boilerplate ():
    print('Optional arguments may be specified in any of three ways:')
    print(' - include them among the positional arguments')
    print(' - in flag form, before the positional arguments')
    print(' - in keyword form, after the positional arguments')
    print()
    print('However, optional arguments that are specified as keyword-only')
    print('cannot be provided positionally.')
    print()
    print('Flags (precede positional arguments):')
    print('    -?         prints this usage message')
    print('    -!         causes a stack trace to be printed in case of error')
    print('    -ARG       value is 1')
    print('    --ARG      value is 1')
    print('    -ARG=VAL')
    print('    --ARG=VAL')
    print('    --         terminates flags')
    print()
    print('Keyword arguments have the form KEY=VAL and follow the positional')
    print('arguments.')


##  Dispatch table.

class _DispatchTable (object):

    ##  Constructor.

    def __init__ (self, table):
        ##  Contents.
        self.table = table
        
    ##  Get value.
    def __getitem__ (self, name):
        if name in self.table:
            return self.table[name]
        else:
            self.error('Unrecognized command: %s' % name)
        
    ##  Print usage and exit.
    def usage (self):
        print('Commands:')
        for name in sorted(self.table):
            print()
            f = _Function(name, self.table[name])
            f.print_description()
        print()
        _print_boilerplate()
        sys.exit(1)
        
    ##  Print error and usage and exit.
    def error (self, msg):
        print('**', msg, file=sys.stderr)
        self.usage()


##  A function.
class _Function (object):

    ##  Constructor.
    def __init__ (self, name, f):

        ##  The name.
        self.name = name

        ##  The python function.
        self.func = None

        ##  Whether it has a 'self' argument.
        self.selfarg = None

        ##  Number of positional parameters.
        self.npos = None

        ##  Number of obligatory parameters.
        self.nobl = None

        ##  Number of optional parameters.
        self.nopt = None

        ##  Number of keyword parameters.
        self.nkwo = None

        ##  Total number of arguments.
        self.nargs = None

        ##  Whether it has a '*args' parameter.
        self.star = None

        ##  Whether it has a '**kwargs' parameter.
        self.starstar = None

        ##  Names of parameters.
        self.names = None

        ##  Default values for parameters.
        self.defaults = None

        ##  Whether it accepts keyword arguments.
        self.keywords = None

        if hasattr(f, '__code__'):
            self.func = f
            self.selfarg = None
        elif hasattr(f, '__call__'):
            m = f.__call__
            if not (hasattr(m, '__self__') and hasattr(m, '__func__')):
                raise Exception('Not a recognized function type')
            self.func = f = m.__func__
            self.selfarg = m.__self__

        self.npos = f.__code__.co_argcount
        if f.__defaults__:
            self.nopt = len(f.__defaults__)
        else:
            self.nopt = 0
        self.nobl = self.npos - self.nopt
        self.nkwo = f.__code__.co_kwonlyargcount
        self.nargs = self.npos + self.nkwo
        self.star = bool(f.__code__.co_flags & 0x04)
        self.starstar = bool(f.__code__.co_flags & 0x08)

        self.names = f.__code__.co_varnames

        if self.selfarg:
            self.npos -= 1
            self.nobl -= 1
            self.nargs -= 1
            self.names = self.names[1:]

        self.defaults = []
        if f.__defaults__: self.defaults.extend(f.__defaults__)
        if self.nkwo:
            for name in self.names[self.npos:self.nargs]:
                self.defaults.append(f.__kwdefaults__[name])

        if self.starstar:
            self.keywords = True
        else:
            self.keywords = set(self.names[self.nobl : self.nargs])

    ##  Whether it accepts kw as a keyword argument.

    def accepts_keyword (self, kw):
        return self.keywords is True or kw in self.keywords

    ##  Print out a description.

    def print_description (self):
        args = ' '.join(self.names[:self.nobl])
        if self.nopt > 0:
            args = '%s [%s]' % (args, ' '.join(self.names[self.nobl:self.npos]))
        print(self.name, args)
        if self.star:
            print('    Additional positional arguments are allowed')
        if self.nkwo > 0:
            print('    Keyword only: ' + ' '.join(self.names[self.npos:self.nargs]))
        if self.defaults:
            print('    Defaults:')
            for i in range(self.nobl, self.nargs):
                print('        %-10s %s' % (self.names[i], self.defaults[i-self.nobl]))
        if self.starstar:
            print('    Additional keyword arguments are allowed')
        if self.func.__doc__:
            print()
            print(self.func.__doc__)

    ##  Print an error message, print usage, and exit.

    def error (self, msg):
        print('**', msg, file=sys.stderr)
        self.usage()
        
    ##  Print usage and exit.

    def usage (self):
        self.print_description()
        print()
        _print_boilerplate()
        sys.exit(1)

    ##  Call the function.

    def __call__ (self, args, kwargs, stack_trace):
        try:
            if self.selfarg: args = [self.selfarg] + args
            value = self.func(*args, **kwargs)
            if value is not None:
                print(value)
            sys.exit(0)
        except Exception as e:
            if stack_trace:
                traceback.print_exc()
            else:
                print('ERROR:', e, file=sys.stderr)
            sys.exit(1)


##  Dispatches either to a DispatchTable or to a Function.

class _Form (object):

    ##  Constructor.  Only one of 'table' or 'function' get set.
    def __init__ (self, main):

        ##  The DispatchTable.
        self.table = None
        ##  The Function.
        self.function = None
        ##  Positional arguments.
        self.args = None
        ##  Keyword arguments.
        self.kwargs = None
        ##  Whether to show a stack trace on error.
        self.stack_trace = None

        if isinstance(main, dict):
            self.table = _DispatchTable(main)
            self.function = None
        else:
            self.table = None
            self.function = _Function(sys.argv[0], main)
        self.args = []
        self.kwargs = {}
        self.stack_trace = False

    ##  The number of optional arguments.
    def n_optional_args (self):
        return self.function.nopt

    ##  The number of obligatory arguments.
    def n_obligatory_args (self):
        return self.function.nobl

    ##  Add arguments.
    def add_args (self, args):
        self.args.extend(args)

    ##  Add a keyword arg.
    def add_keyword_arg (self, key, value):
        if self.function.accepts_keyword(key):
            if key in self.kwargs:
                self.error('Keyword specified multiple times: %s' % key)
            self.kwargs[key] = value
        else:
            self.error('Unrecognized keyword: %s' % key)

    ##  Call the function on arguments and keyword arguments.
    def eval (self):
        self.function(self.args, self.kwargs, self.stack_trace)

    ##  Print an error message, print usage, and exit.
    def error (self, msg):
        if self.function: self.function.error(msg)
        else: self.table.error(msg)

    ##  Print usage and exit.
    def usage (self):
        if self.function: self.function.usage()
        else: self.table.usage()


def _isflag (i):
    return i < len(sys.argv) and sys.argv[i].startswith('-') and sys.argv[i] != '-'

def _isnonflag (i):
    return i < len(sys.argv) and ((not sys.argv[i].startswith('-')) or sys.argv[i] == '-')


##  Deprecated.  A command-line parser.
class _CommandLineParser (object):

    ##  Constructor.

    def __init__ (self):

        ##  Beginning of obligatory args.
        self.i = None
        ##  Beginning of optional args.
        self.j = None
        ##  Beginning of keyword args.
        self.k = None

    ##  Call it.
    def __call__ (self, main):
        form = _Form(main)
        self.i = 1
        while self.special_flag(form): pass
        if form.function is None: self.command(form)
        self.flags(form)
        self.obligatory_args(form)
        self.keyword_args(form)
        self.optional_args(form)
        return form

    ##  Whether a special flag has been provided (-? or -!).
    def special_flag (self, form):
        if self.i >= len(sys.argv): return False
        arg = sys.argv[self.i]
        if arg == '-!': form.stack_trace = True
        elif arg == '-?': form.usage()
        else: return False
        self.i += 1
        return True

    ##  Set the command.
    def command (self, form):
        self.i = 1
        if _isnonflag(self.i):
            name = sys.argv[self.i]
            com_name = '%s %s' % (sys.argv[0], name)
            self.i += 1
            form.function = _Function(name, form.table[name])
        else:
            form.error('No command provided')

    ##  Process flags.
    def flags (self, form):
        while _isflag(self.i):
            if not self.special_flag(form):
                arg = sys.argv[self.i]
                self.i += 1
                if arg == '--': break
                t = arg.find('=')
                if t < 0:
                    flag = arg[1:]
                    value = '1'
                else:
                    u = 1
                    if arg[u] == '-': u += 1
                    flag = arg[u:t]
                    value = arg[t+1:]
                form.add_keyword_arg(flag, value)

    ##  Process obligatory args.
    def obligatory_args (self, form):
        self.j = self.i + form.n_obligatory_args()
        if self.j > len(sys.argv):
            form.error('Not enough arguments provided')
        form.add_args(sys.argv[self.i:self.j])

    ##  Process keyword args.
    def keyword_args (self, form):
        self.k = len(sys.argv)
        while self.k > self.j:
            arg = sys.argv[self.k-1]
            t = arg.find('=')
            if t < 0: break
            self.k -= 1
            (key,value) = (arg[:t], arg[t+1:])
            form.add_keyword_arg(key, value)
        
    ##  Process optional args.
    def optional_args (self, form):
        if self.k - self.j > form.n_optional_args():
            form.error('Too many arguments provided')
        form.add_args(sys.argv[self.j:self.k])


_command_line = None

##  Deprecated.  Use shift instead.

def error (msg=None):
    global _command_line
    if _command_line is None:
        raise Exception('Cannot be called except in the context of run')
    _command_line.error(msg)

##  Deprecated.  Use shift instead.

def run (main):
    global _command_line
    p = _CommandLineParser()
    _command_line = p(main)
    _command_line.eval()


#--  Command line processing, v5  ----------------------------------------------

##  Process command-line arguments.

class Shift (object):

    ##  Constructor.  The argv should not include the command, only the arguments.

    def __init__ (self, argv=sys.argv, offset=1):
        self._usage = ''

        ##  The command line.
        self.argv = argv

        ##  Current pointer into argv.
        self.ac = 0

        self._initial_offset = offset

    ##  Set the usage message.

    def set_usage (self, msg):
        if len(self.argv) > 1 and self.argv[1] in ('-?', '--help'):
            print(msg)
            sys.exit(0)
        self._usage = msg

    ##  Print usage.

    def print_usage (self):
        if self._usage:
            print('USAGE:', file=sys.stderr)
            print(self._usage, file=sys.stderr)
        else:
            print('No usage information available', file=sys.stderr)

    ##  Begin processing.

    def __enter__ (self):
        self.ac = self._initial_offset
        return self

    ##  Calls done().

    def __exit__ (self, t, b, tb):
        if t is None:
            self.done()

    ##  Print out an error message, including usage, and quit.

    def error (self, msg):
        print('**', msg, file=sys.stderr)
        self.print_usage()
        sys.exit(1)

    ##  Whether any more arguments remain.

    def peek (self, targ=None):
        if targ is None:
            if self.ac < len(self.argv):
                return self.argv[self.ac]
        else:
            return (self.ac < len(self.argv) and self.argv[self.ac] == targ)

    ##  Whether the next argument is a flag.

    def isflag (self):
        return self.ac < len(self.argv) and self.argv[self.ac].startswith('-')

    ##  If the next argument is a flag, return it and advance.

    def flag (self):
        if self.isflag():
            return self()

    ##  Return the next argument and advance.

    def __call__ (self, targ=None):
        if targ is None:
            if self.ac >= len(self.argv):
                self.error('Too few arguments')
            arg = self.argv[self.ac]
            self.ac += 1
            return arg
        elif self.ac < len(self.argv) and self.argv[self.ac] == targ:
            self.ac += 1
            return True
        else:
            return False

    ##  Return the next argument and advance, if there is a next argument.

    def ifable (self):
        if self.ac < len(self.argv):
            arg = self.argv[self.ac]
            self.ac += 1
            return arg

    ##  Returns True just in case __call__ will succeed.

    def able (self):
        return self.ac < len(self.argv)

    ##  Whether or not we have processed all arguments.

    def isdone (self):
        return self.ac >= len(self.argv)

    ##  Return the remaining arguments as a list, and advance to the end.

    def rest (self):
        args = self.argv[self.ac:]
        self.ac = len(self.argv)
        return args
    
    ##  Check whether all arguments have been consumed.  Signal an error if not.

    def done (self):
        if self.ac != len(self.argv):
            self.error('Too many arguments')


##  An instance of Shift reading sys.argv[1:].
shift = Shift(sys.argv[1:])


#--  Main  ---------------------------------------------------------------------

def _docstring_lines (x):
    doc = x.__doc__
    if doc:
        if doc.startswith('\n'):
            i = 1
            while i < len(doc) and doc[i].isspace() and doc[i] not in '\r\n':
                i += 1
            prefix = doc[1:i]
            doc = doc[i:]
        else:
            prefix = ''
        n = len(prefix)
        for line in doc.split('\n'):
            if line.startswith(prefix):
                line = line[n:]
            yield line


class Main (object):

    def _usage_message (self):
        yield 'Usage: COMMAND [-FLAG[=VAL]*] [ARG*]'

        if self.__doc__:
            yield ''
            for line in _docstring_lines(self):
                yield line

        yield ''
        yield 'Commands:'
        for name in dir(self):
            if name.startswith('com_'):
                yield ''
                com = name[4:]
                method = getattr(self, name)
                dflts = method.__defaults__ or []
                nkws = len(dflts)
                varnames = method.__code__.co_varnames
                nargs = method.__code__.co_argcount - (1 + nkws)  # -1 for self
                i = 1          # first is self
                j = 1 + nargs
                k = j + nkws   # after kws are local variables
                args = varnames[i:j]
                kws = varnames[j:k]
                words = [com]

                # print('** name=', name)
                # print('   dflts=', dflts, 'varnames=', varnames, 'nargs=', nargs, 'nkws=', nkws)
                # print('   kws=', kws, 'args=', args, 'i=', i, 'j=', j, 'k=', k)

                words.extend(args)
                yield '  ' + ' '.join(words)

                if kws:
                    for (i, kw) in enumerate(kws):
                        yield '      -' + kw + ' default: ' + repr(dflts[i])
                
                if method.__doc__:
                    for line in _docstring_lines(method):
                        yield '    ' + line

    def __call__ (self, comline=None):
        if comline is None:
            args = sys.argv
        else:
            args = [None] + comline.split()

        with Shift(args) as shift:
            shift.set_usage('\n'.join(self._usage_message()))
            args = []
            while not (shift.isdone() or shift.isflag()):
                args.append(shift())
            kwargs = {}
            while shift.isflag():
                flag = shift()
                key = flag[1:]
                i = key.find('=')
                value = True
                if i >= 0:
                    value = key[i+1:]
                    key = key[:i]
                kwargs[key] = value
            args.extend(shift.rest())
    
        com = None
        nwords = 0
        if not args:
            print('** No command given')
            sys.exit(1)
        for n in range(1, len(args)+1):
            methodname = 'com_' + '_'.join(args[:n])
            if hasattr(self, methodname):
                com = getattr(self, methodname)
                nwords = n
        if com is None:
            print('** Not a command:', args[0])
            sys.exit(1)
        args = args[nwords:]
        com(*args, **kwargs)


#--  Timeout  ------------------------------------------------------------------

##  Exception used when a timeout occurs.

class TimedOut (Exception): pass

##  Runs a timer.

class Timeout (object):

    ##  Constructor.

    def __init__ (self, nsecs):

        ##  How long before the alarm goes off.
        self.nsecs = nsecs

        ##  The timer, in another thread, while running.
        self.timer = None

    ##  Enter.  Starts a timer in a thread.  If the timer goes off before
    #   it is canceled, it raises a TimedOut exception.

    def __enter__ (self):
        self.timer = threading.Timer(self.nsecs, self._alarm)

    def _alarm (self):
        raise TimedOut()

    ##  Exit.  Cancel the timer.  If TimedOut is raised, pass it through.

    def __exit__ (self, t, v, tb):
        self.timer.cancel()
        if t == TimedOut:
            return True


# class Timeout (object):
# 
#     def __init__ (self, nsecs, ontimeout=None):
#         self.nsecs = nsecs
#         self.ontimeout = None
# 
#     def __enter__ (self):
#         signal.signal(signal.SIGALRM, self._alarm)
#         signal.alarm(self.nsecs)
# 
#     def _alarm (self, signo, frame):
#         raise TimedOut()
# 
#     def __exit__ (self, t, v, tb):
#         signal.alarm(0)
#         signal.signal(signal.SIGALRM, signal.SIG_DFL)
#         if t == TimedOut:
#             return True

#  This works, but very heavyweight and clunky!!
#  Requires one to communicate with body via files or pipes!!
#
# class with_timeout (object):
# 
#     def __init__ (self, nsecs, fnc, *args, **kwargs):
#         self.childpid = None
#         self.timedout = False
# 
#         self.childpid = os.fork()
# 
#         # child
#         if self.childpid == 0:
#             fnc(*args, **kwargs)
#             os.kill(os.getpid(), SIGTERM)
# 
#         # parent
#         else:
#             timer = threading.Timer(nsecs, self._alarm)
#             timer.start()
#             (pid, status) = os.wait()
#             if not self.timedout:
#                 timer.cancel()
# 
#     def _alarm (self):
#         os.kill(self.childpid, SIGTERM)
#         self.timedout = True


#  Threading doesn't work.  Threads can't be interrupted, and even the
#  main thread doesn't always response to interrupt_main().
# 
# def _timer_goes_off ():
#     from seal.app.log import DEBUG
#     DEBUG('timer goes off')
#     interrupt_main()
# 
# class Timeout (object):
# 
#     def __init__ (self, nsecs, ontimeout=None):
#         self.timer = None
#         self.nsecs = nsecs
#         self.ontimeout = ontimeout
# 
#     def __enter__ (self):
#         self.timer = threading.Timer(self.nsecs, _timer_goes_off)
#         from seal.app.log import DEBUG
#         DEBUG('** starting timer, nsecs=', self.nsecs, 'thread=', repr(threading.current_thread()))
#         self.timer.start()
# 
#     def __exit__ (self, t, v, tb):
#         from seal.app.log import DEBUG
#         DEBUG('** cancelling timer', 't=', t)
#         self.timer.cancel()
#         if t == KeyboardInterrupt:
#             DEBUG('** got an interrupt')
#             if self.ontimeout is not None:
#                 self.ontimeout()
#             # suppress the exception
#             return True


#--  Elapsed Time  -------------------------------------------------------------

##  Prints out as elapsed time since it was created.

class Timer (object):

    ##  Constructor.

    def __init__ (self):

        ##  The start time.
        self.start = time()

    def elapsed (self):
        return time() - self.start

    ##  Prints elapsed time.

    def __str__ (self):
        return elapsed_time_str(self.start, time())


#--  Progress Monitor  ---------------------------------------------------------

##  A progress monitor.

class Progress (object):

    ##  Constructor.

    def __init__ (self, n=None):

        ##  How many ticks are expected in total.  It's OK if it's an underestimate.
        self.target = n

        ##  How many ticks have already taken place.
        self.count = 0

        ##  To keep track of elapsed time.
        self.timer = Timer()

        self.last_t = None

    def __enter__ (self):
        return self

    def done (self):
        self.printout(end='\n')

    def __exit__ (self, t, v, tb):
        self.done()

    ##  Increment by n ticks (default 1).  Prints/updates a progress message.

    def __iadd__ (self, n=1):
        self.count += n
        t = time()
        if self.last_t is None or t - self.last_t > 0.3:
            self.last_t = t
            self.printout()
        return self

    def printout (self, end=' '):
        if self.target is None:
            print('\rProgress: %d' % self.count,
                  'Time elapsed: %s' % self.timer,
                  end=end,
                  file=sys.stderr)
            sys.stderr.flush()

        else:
            proportion_done = self.count/float(self.target)
            elapsed = self.timer.elapsed()
            est_total = elapsed/proportion_done
            print('\rProgress: %d/%d (%2.2f%%)' % (self.count, self.target, 100 * proportion_done),
                  'Time remaining: %s' % elapsed_time_str(elapsed, est_total),
                  end=end,
                  file=sys.stderr)
            sys.stderr.flush()


#--  Load module  --------------------------------------------------------------

##  Load a module, given a fully qualified name.

def load_module (name):
    names = name.split('.')
    module = file = path = None
    try:
        for name in names:
            (file, path, desc) = imp.find_module(name, path)
            module = imp.load_module(name, file, path, desc)
            if file is not None: file.close()
            if hasattr(module, '__path__'): path = module.__path__
    finally:
        if file is not None: file.close()
    return module

##  Import a class, given a fully qualified name.

def import_class (spec):
    i = spec.rfind('.')
    if i < 0: raise Exception('Need a fully qualified class name: %s' % spec)
    name = spec[i+1:]
    modname = spec[:i]
    module = importlib.import_module(modname)
    try:
        cls = module.__dict__[name]
    except KeyError:
        raise Exception('No class named %s in %s' % (name, modname))
    return cls

##  The fully qualified name for a given class.

def class_name (cls):
    return cls.__module__ + '.' + cls.__name__


#--  Manifest  -----------------------------------------------------------------
#
#  A manifest is a listing of the files in a directory tree.  The entries are of
#  form:
#
#      name relpath size [modtime] [hash]
#
#  Options.  By default, all are False:
#
#      symlinks=True:  Follow symlinks.  Signal an error for dangling symlinks
#                      or symlink chains.  If symlinks=False, the entry for a
#                      symlink has the target's name in place of size.
#
#      tmpfiles=True:  Include tmp files.  Tmp file patterns: *~ #*# tmp tmp.* *.tmp
#
#      downcase=True:  Downcase all names.  I.e., case-insensitive matching.
#
#      modtimes=True:  Include modtimes in the output.
#
#      hashes=True:    Include SHA5 hash values in the output.
#

# def manifest (dir, symlinks=False, tmpfiles=False, downcase=False, hashes=False):
#     return sorted(iter_tree(dir, symlinks, tmpfiles))
# 
# def iter_tree (dir, symlinks=False, tmpfiles=False, downcase=False, hashes=False):
#     names = os.listdir(dir)
#     for name in names:
#         fn = os.path.join(dir, name)
#         islink = False
#         size = None
#         modtime = None
#         hash = None
#         if os.path.islink(fn):
#             islink = True
#             tgt = os.readlink(fn)
#             if symlinks:
#                 if not os.path.exists(tgt):
#                     raise Exception('Dangling symlink: %s -> %s' % (fn, tgt))
#                 if os.path.islink(tgt):
#                     raise Exception('Symlink chain: %s -> %s' % (fn, tgt))
#                 fn = tgt
#                 islink = False
#             else:
#                 size = tgt
#         if not islink:
#             if os.path.isdir(fn):
#                 for entry in iter_tree(fn):
#                     yield entry
#                 continue
#             else:
#                 size = file_size(fn)
#                 


#--  XTerm  --------------------------------------------------------------------

##  Dict mapping color names to escape strings.
fg = {'black': '\033[30m',
      'red': '\033[31m',
      'green': '\033[32m',
      'yellow': '\033[33m',
      'blue': '\033[34m',
      'magenta': '\033[35m',
      'cyan': '\033[36m',
      'white': '\033[37m',
      'default': '\033[39m'}

##  Dict mapping color names to escape strings.
bg = {'black': '\033[40m',
      'red': '\033[41m',
      'green': '\033[42m',
      'yellow': '\033[43m',
      'blue': '\033[44m',
      'magenta': '\033[45m',
      'cyan': '\033[46m',
      'white': '\033[47m',
      'default': '\033[49m'}

##  Returns a string that displays the contents in black.
def black (s):
    return fg['black'] + s + fg['default']

##  Returns a string that displays the contents in red.
def red (s):
    return fg['red'] + s + fg['default']

##  Returns a string that displays the contents in green.
def green (s):
    return fg['green'] + s + fg['default']

##  Returns a string that displays the contents in yellow.
def yellow (s):
    return fg['yellow'] + s + fg['default']

##  Returns a string that displays the contents in blue.
def blue (s):
    return fg['blue'] + s + fg['default']

##  Returns a string that displays the contents in magenta.
def magenta (s):
    return fg['magenta'] + s + fg['default']

##  Returns a string that displays the contents in cyan.
def cyan (s):
    return fg['cyan'] + s + fg['default']

##  Returns a string that displays the contents in white.
def white (s):
    return fg['white'] + s + fg['default']

##  An escape string that moves the cursor n spaces to the right.
def cursor_right (n=1):
    return '\033%dC' % n

##  An escape string that moves the cursor n spaces to the left.
def cursor_left (n=1):
    return '\033%dD' % n

##  An escape string that moves the cursor to column n.
def goto_column (n):
    return '\033%dG' % n




# Temporary solution until Python 3.9 becomes standard

from os import PathLike
from os.path import join, dirname

class _Path (PathLike):

    def __init__ (self, fn):
        self._filename = fn
        
    def __enter__ (self):
        return self

    def __exit__ (self, t, v, tb):
        return

    def __add__ (self, suffix):
        return _Path(self._filename + suffix)

    def __str__ (self):
        return self._filename

    def __repr__ (self):
        return f'<_Path {self._filename}>'

    def __fspath__ (self):
        return self._filename
    

def path (*names):
    return _Path(join(dirname(__file__), *names))

def ex (*names):
    return path('examples', *names)



##  @package seal.core.io
#   The io module contains input and output streams.

try:
    from fcntl import flock, LOCK_EX, LOCK_UN
except:
    flock = LOCK_EX = LOCK_UN = None

import sys, os, re, urllib.request, urllib.parse, urllib.error, \
    codecs, shutil, subprocess, stat, threading, datetime, random, \
    json
from glob import glob
from io import StringIO, TextIOBase
from os.path import expanduser
from pathlib import Path
from .seq import as_list
from .string import as_ascii
from .com import fg

_open = __builtins__['open']


def ispathlike (x):
    return isinstance(x, str) or hasattr(x, '__fspath__')


##  Passes '-' through unchanged; otherwise dispatches to os.path.abspath().

def abspath (fn):
    if fn == '-':
        return fn
    else:
        return os.path.abspath(os.path.expanduser(fn))


#--  srepr  --------------------------------------------------------------------

##  'Stable' repr, for use with doctest.
#   E.g., prints out the keys of a dict or members of a set in sort order.

def srepr (x):
    if isinstance(x, dict):
        pairs = ['%s: %s' % (repr(k), srepr(v))
                 for (k,v) in sorted(x.items())]
        return '{' + ', '.join(pairs) + '}'

    elif isinstance(x, set):
        elts = [srepr(e) for e in sorted(x)]
        return '{' + ', '.join(elts) + '}'

    else:
        return repr(x)


#--  Contents  -----------------------------------------------------------------

##  The contents of a file, as a string.

def contents (filename, encoding=None):
    with infile(filename, encoding) as f:
        return f.read()


#--  Tee  ----------------------------------------------------------------------

##  A tee.  Splits a stream into one going to a file and an identical copy going
#   to stdout.

class tee (object):

    ##  Constructor.

    def __init__ (self, filename):

        ##  The filename being written.
        self.filename = filename

        ##  The output stream to the file.
        self.stream = _open(filename, 'w')

    ##  Writes the given string both to the output stream and to stdout.

    def write (self, x):
        self.stream.write(x)
        sys.stdout.write(x)

    ##  Closes the output stream.

    def close (self):
        self.stream.close()
        self.stream = None
    

#--  NullStream  ---------------------------------------------------------------

##  A fake stream.

class NullStream (object):
    
    ##  No-op.

    def write (self, s):
        pass

##  An instance of NullStream.
null = NullStream()


#--  Output List  --------------------------------------------------------------

##  A specialization of list that behaves like an output stream.
#   Specifically, it provides a write() method.  Writing to it appends lines;
#   the final result is a list of lines.  The lines do not include carriage returns
#   and newlines.

class OutputList (list):

    ##  Constructor.

    def __init__ (self):

        ##  A partial line, still waiting for the newline.
        self.partial = None

    ##  Write a string.  If the string ends in carriage return and/or newline,
    #   it completes a line.  Append it to the list and clear the partial member.
    #   Otherwise, concatenate it to the partial member and wait.
    #   Caution: does not handled embedded newlines.

    def write (self, s):
        eol = False
        if s and s[-1] in '\r\n':
            eol = True
            s = s.rstrip('\r\n')
        if self.partial:
            self.partial += s
        else:
            self.partial = s
        if eol:
            self.append(self.partial)
            self.partial = None


#--  Suffixes  -----------------------------------------------------------------

##  A suffix begins with the final period, provided that it occurs in the final
#   pathname component.  Returns the pathname sans suffix.

def strip_suffix (fn):
    i = fn.rfind('.')
    if i > 0 and '/' not in fn[i+1:]:
        return fn[:i]
    else:
        return fn

##  Returns a pair: (filename sans suffix, suffix sans period).

def split_suffix (fn):
    i = fn.rfind('.')
    if i > 0 and '/' not in fn[i+1:]:
        return (fn[:i], fn[i+1:])
    else:
        return (fn, '')

##  Returns the suffix or an empty string.  The period is not included.

def get_suffix (fn):
    i = fn.rfind('.')
    if i > 0 and '/' not in fn[i+1:]:
        return fn[i+1:]
    else:
        return ''


#--  Fn  -----------------------------------------------------------------------

##  The only purpose of this object is so that we can distinguish by type
#   between a filename and the contents of a file.

class Contents (object):

    ##  Constructor.

    def __init__ (self, s):

        ##  The string.
        self.contents = s


##  A filename; a specialization of str.

class Fn (str):

    ##  Dot can be used to add a pathname component.

    def __getattr__ (self, attr):
        if os.path.isdir(self):
            return Fn(os.path.join(self, attr))
        else:
            return Fn(self + '.' + attr)

    ##  Plus can also be used to concatenate a string; no slash is inserted.

    def __add__ (self, other):
        return Fn(str.__add__(self, other))

    ##  Divide can be used to add pathname components.

    def __div__ (self, other):
        return self.join(other)

    ##  Double slash can also be used.

    def __truediv__ (self, other):
        return self.join(other)

    ##  CAUTION: this overrides str.join, with a very different semantics.
    #   Specifically, this does a pathname join.

    def join (self, *cpts):
        return Fn(os.path.join(self, *cpts))

    ##  Whether the named file exists.

    def exists (self):
        return os.path.exists(self)

    ##  True if the named file exists and is a directory.

    def isdir (self):
        return os.path.isdir(self)

    ##  The parent, obtained by stripping off the last component and deleting
    #   any trailing slashes.

    def parent (self):
        if self.endswith('/'):
            name = self.rstrip('/')
        else:
            name = self
        if name:
            return Fn(os.path.dirname(name))

    ##  Create the file.  Create any missing ancestor using os.makedirs(),
    #   then touch the file.

    def create (self, dir=False):
        if self.endswith('/'): dir = True
        if dir:
            os.makedirs(self)
        else:
            dirname = os.path.dirname(self)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            _open(self, 'w').close()

    ##  List the named file.  No-op if the file does not exist.  Returns
    #   a list of child file names, if it names an existing directory.
    #   Returns the result of os.stat(), if it names an existing regular file.

    def list (self):
        if os.path.exists(self):
            if os.path.isdir(self):
                return [Fn(f) for f in os.listdir(self)]
            else:
                return os.stat(self)

    ##  Dispatches to shutil.move().

    def rename (self, newname):
        shutil.move(self, newname)

    ##  Makes a copy.  If this is a directory, it does a deep copy.

    def copy (self, newname, symlinks=True):
        if os.path.isdir(self):
            shutil.copytree(self, newname, symlinks)
        else:
            shutil.copyfile(self, newname)

    ##  Delete the file.

    def delete (self):
        if os.path.isdir(self):
            os.rmdir(self)
        else:
            os.unlink(self)

    ##  Return the file size.

    def size (self):
        return file_size(self)

    ##  Return the modtime, in seconds since the epoch (midnight UTC, Jan 1, 1970).

    def modtime (self):
        return os.stat(self).st_mtime


# ##  The Seal python directory: the inner 'seal' directory.
# __seal__ = os.path.dirname(os.path.dirname(__file__))
# 
# ##  The Seal destination directory: the outer 'seal' directory.
# __dest__ = os.path.dirname(os.path.dirname(__seal__))
# 
# ##  The Seal destination directory.
# dest = Fn(__dest__)
# 
# ##  The Seal bin directory.
# bin = Fn(os.path.join(__dest__, 'bin'))
# 
# ##  The seal examples directory.
# ex = Fn(os.path.join(__seal__, '__examples__'))
# 
# ##  The seal data directory.
# data = Fn(os.path.join(__seal__, '__data__'))
# 
# ##  The root directory.
# root = Fn(os.path.sep)
# 
# ##  The tmp directory.
# tmp = root.tmp
# 
# ##  The current directory, as an Fn.
# here = Fn(os.path.curdir)
# 
# ##  The user's home directory, as an Fn.
# home = Fn(os.path.expanduser('~'))
# 
# ##  Nowhere (/dev/null).
# nowhere = Fn('/dev/null')


##  Allocate a tmp filename.  It automatically gets deleted when
#   the filename is deleted or garbage collected.
#   That is, one may delete the file by calling del on the filename.

def tmpfile ():
    _TmpFn.count += 1
    fn = _TmpFn('/tmp/seal_%d_%d' % (os.getpid(), _TmpFn.count))
    fn._delete_on_gc = True
    return fn

##  Allocates a temporary filename prefix and arranges for any file whose
#   name begins with that prefix to be
#   deleted (if any such exist) when the C{TmpFile} object gets garbage
#   collected.  The filename prefix is of form::
#
#       /tmp/seal_PID_N_
#
#   where PID is the pid of the python process and N is a
#   sequence number that is advanced each time a temporary filename is
#   allocated.

class _TmpFn (Fn):

    ##  So that we assign a unique number to each temp file.
    count = 0

    ##  When the TmpFn gets deleted, delete the file on disk.
    def __del__ (self):
        if self._delete_on_gc:
            os.system("rm -f " + self + "*")

    ##  Call this to prevent file deletion on garbage collection.
    def no_delete (self):
        self._delete_on_gc = False


#--  infile and outfile  -------------------------------------------------------

##  A readable representation; fn may be a filename, a StringIO, or a stream.

def filename_string (fn):
    if isinstance(fn, str):
        return fn
    elif isinstance(fn, StringIO):
        return '<string input>'
    else:
        return '<open file>'

##  Open a file for reading.  Filename may be '-', a string representing a
#   filename, a StringIO, a stream, or anything that has a readline() method.

def infile (filename, encoding=None):
    if isinstance(filename, str):
        if not encoding: encoding = 'utf8'
        if filename == '-':
            return sys.stdin
        elif re.match(r'[A-Za-z]+:', filename):
            bstream = urllib.request.urlopen(filename, 'r')
            reader = codecs.getreader(encoding)
            return reader(bstream)
        else:
            return _open(filename, 'r', encoding=encoding)
    elif hasattr(filename, '__fspath__'):
        return infile(filename.__fspath__())
    elif isinstance(filename, StringIO):
        return filename
    elif isinstance(filename, TextIOBase):
        return filename
    elif hasattr(filename, 'readline'):
        return filename
    else:
        raise Exception('Cannot coerce to file: %s' % filename)

##  Open a file for writing.  Fileid may be '-', a string representing
#   a filename, or None.  If None, a StringIO is returned.

def outfile (fileid=None, encoding=None):
    if fileid is None:
        return StringIO()
    elif fileid == '-':
        return sys.stdout
    else:
        if not encoding: encoding = 'utf8'
        return _open(fileid, 'w', encoding=encoding)

##  An infile containing bytes.

def byte_infile (fn):
    if fn == '-': return sys.stdin.buffer
    else: return _open(fn, 'rb')

##  An outfile writing bytes.

def byte_outfile (fn):
    if fn == '-': return sys.stdout.buffer
    else: return _open(fn, 'wb')

##  Close a file, but not if it is a StringIO or a std stream.

def close (file):
    if isinstance(file, StringIO):
        value = file.getvalue()
    else:
        value = None
    if file not in [sys.stdin, sys.stdout, sys.stderr]:
        file.close()
    return value


#--  Load/save string  ---------------------------------------------------------

##  Load a string from a file.

def load_string (filename, encoding=None):
    f = infile(filename, encoding)
    return f.read()

##  Save a string to a file.

def save_string (s, filename=None, encoding=None):
    f = outfile(filename, encoding)
    f.write(s)
    return close(f)


#--  Lines  --------------------------------------------------------------------

##  Load a list of lines from a file.  Newlines and carriage returns are not
#   included in the lines.

def load_lines (filename, encoding=None):
    return list(iter_lines(filename, encoding))

##  Iterate over a list of lines.

def iter_lines (filename, encoding=None):
    return read_lines(infile(filename, encoding))

def read_lines (f):
    for line in f:
        yield line.rstrip('\r\n')

##  Save a list of lines to a file.

def save_lines (lines, filename=None, encoding=None):
    write_lines(lines, outfile(filename, encoding))

def write_lines (lines, f):
    for line in lines:
        f.write(line)
        f.write('\n')
    return close(f)


#--  Records  ------------------------------------------------------------------

##  Load tabular records from a file.

def load_records (filename, header=None, separator='\t', encoding=None):
    return list(iter_records(filename, header, separator, encoding))

##  Iterate over tabular records.

class iter_records (object):

    ##  Constructor.

    def __init__ (self, filename, header=None, separator='\t', encoding=None):
        ##  Filename.
        self.filename = filename_string(filename)
        ##  File.
        self.file = infile(filename, encoding)
        ##  Number of lines processed so far.
        self.linecount = 0
        ##  Separator.
        self.separator = separator
        if header is not None:
            line = self.__readline()
            fileheader = line.rstrip('\r\n').split(separator)
            if fileheader != header:
                raise Exception("Bad header: got %s, expecting %s: %s" % \
                    (fileheader, header, filename))

    def __readline (self):
        self.linecount += 1
        return next(self.file)

    ##  Returns self.

    def __iter__ (self):
        return self

    ##  The iteration method.

    def __next__ (self):
        line = self.__readline()
        return line.rstrip('\r\n').split(self.separator)

    ##  Signal an error, adding the filename and line number.

    def error (self, msg):
        raise Exception("[%s line %d] %s" % (self.filename, self.linecount, msg))

##  Save a list of records to a tabular file.

def save_records (records, filename=None, header=None, separator='\t', encoding=None):
    f = outfile(filename, encoding)
    if header:
        f.write(separator.join(header))
        f.write('\n')
    for record in records:
        f.write(separator.join(record))
        f.write('\n')
    return close(f)


#--  Dict  ---------------------------------------------------------------------

##  Load a dict from a file.

def load_dict (filename, encoding=None):
    return dict((k,v) for (k,v) in load_records(filename, encoding))

##  Save a dict to a file.  The output is tabular, with keys in the first column
#   and values in the second column.

def save_dict (d, filename=None, encoding=None):
    return save_records(iter(d.items()), filename, encoding)

    
#--  NestedDict  ---------------------------------------------------------------

##  Load a nested dict from a file.
#   A nested dict permits keys containing dots.  The first part of the key
#   becomes a key in the toplevel dict, and its value is a subdict.

def load_nested_dict (fn, encoding=None):
    out = {}    
    lines = list(_open(fn, encoding=encoding))
    lno = 0
    while lno < len(lines):
        line = lines[lno].rstrip('\r\n')
        if (not line) or line.startswith('#'): continue
        (keys, v) = _nd_split(lno, line)
        (d, k) = _nd_follow_path(lno, out, keys)
        lno += 1
        if not v:
            (v, lno) = _nd_list_value(lines, lno)
        d[k] = v
    return out

def _nd_split (lno, line):
    i = 0
    while i < len(line) and not line[i].isspace():
        i += 1
    keys = line[:i].split('.')
    v = line[i:].strip()
    return (keys, v)

def _nd_follow_path (lno, d, keys):
    for k in range(len(keys)-1):
        key = keys[k]
        if key in d:
            d = d[key]
            if not isinstance(d, dict):
                s = '.'.join(keys[:k+1])
                raise Exception(f'[line {lno}] Value for {s} is not a dict')
        else:
            newd = {}
            d[key] = newd
            d = newd
    return (d, keys[-1])

def _nd_list_value (lines, lno):
    vals = []
    while lno < len(lines) and len(lines[lno]) > 0 and lines[lno][0].isspace():
        vals.append(lines[lno].strip())
        lno += 1
    return (vals, lno)
        

##  Write a dict in nested dict format.

def save_nested_dict (d, fn, encoding=None):
    f = outfile(fn, encoding)
    _write_nested_dict('', d, f)
    return close(f)

def _write_nested_dict (prefix, d, f):
    for key in d:
        if prefix: fullkey = prefix + '.' + key
        else: fullkey = key
        v = d[key]
        if isinstance(v, dict):
            _write_nested_dict(fullkey, v, f)
        else:
            f.write(fullkey)
            f.write(' ')
            v = str(v)
            if '\n' in v:
                raise Exception('Value contains newline: %s' % repr(v))
            f.write(v)
            f.write('\n')


#--  Config file  --------------------------------------------------------------

#
#  [0][d1]
#      (0, 'server', None)
#  [0 4][d1 d2]
#      (4, 'cld', None)
#  [0 4 8][d1 d2 d3]
#      (8, 'foo', '10')
#      (8, 'bar', '2')
#      (4, 'baz', '6')
#


##  Produces a nested dict, but instead of dotted keys, it uses indentation
#   Keys and values are separated by '='.
#   Hash starts a comment line.
#   Empty lines are ignored.
#
#   Example:
#
#       server
#           cld
#               foo = 10
#               bar = 2
#           baz = 6

def load_config (fn, encoding=None):
    lines = list(_iter_config_lines(fn, encoding))
    cfg = {}
    indents = [0]
    dicts = [cfg]

    for i in range(len(lines)):
        (ind, key, value) = lines[i]

        # Pop as necessary
        if ind > indents[-1]:
            raise Exception('Bad indentation')
        while ind < indents[-1]:
            if len(indents) == 1:
                raise Exception('Bad indentation in config file')
            indents.pop()
            dicts.pop()
        if ind != indents[-1]:
            raise Exception('Bad indentation in config file')            

        # Start subdict
        if value is None:
            if i+1 >= len(lines):
                raise Exception("No body for subdict '%s'" % key)
            nextind = lines[i+1][0]
            if nextind <= ind:
                raise Exception("No body for subdict '%s'" % key)
            d = {}
            dicts[-1][key] = d
            indents.append(nextind)
            dicts.append(d)

        # Set value
        else:
            dicts[-1][key] = value

    return cfg

def _iter_config_lines (fn, encoding):
    for line in infile(fn, encoding):
        line = line.rstrip('\r\n')
        if (not line) or line.startswith('#'): continue
        ind = 0
        while ind < len(line) and line[ind].isspace():
            if line[ind] != ' ':
                raise Exception('Only spaces permitted in config file')
            ind += 1
        print('ind=', ind)
        i = ind
        while True:
            if i >= len(line):
                key = line.strip()
                value = None
                break
            elif line[i] == '=':
                key = line[ind:i].strip()
                value = line[i+1:].strip()
                break
            else:
                i += 1
        yield (ind, key, value)


#--  Paragraphs  ---------------------------------------------------------------

##  Load a list of paragraphs from a file.  In the file, paragraphs are separated
#   by empty lines.  In memory, paragraphs are represented as single strings with
#   embedded newlines.

def load_paragraphs (filename, encoding=None):
    return list(iter_paragraphs(filename, encoding))

##  Iterate over paragraphs.

def iter_paragraphs (filename, encoding=None):
    file = infile(filename, encoding)
    par = []
    try:
        while True:
            line = next(file)
            if line == '\n':
                yield ''.join(par)
                par = []
            else:
                par.append(line)
    except StopIteration:
        if par:
            yield ''.join(par)

##  Write out a list of paragraphs to a file, inserting empty lines between them.

def save_paragraphs (pars, filename=None, encoding=None):
    file = outfile(filename, encoding)
    first = True
    for par in pars:
        if first:
            first = False
        else:
            file.write('\n')
        file.write(par)
    return close(file)


#--  Blocks  -------------------------------------------------------------------

##  Load a list of blocks from a file.  A block is a list of lines, sans newlines.
#   In the file, blocks are separated by empty lines.

def load_blocks (filename, encoding=None):
    return list(iter_blocks(filename, encoding))

##  Iterate over blocks.

def iter_blocks (filename, encoding=None):
    return read_blocks(infile(filename, encoding))

def read_blocks (f):
    block = []
    for line in read_lines(f):
        if line:
            block.append(line)
        else:
            if block:
                yield block
                block = []
    if block:
        yield block

##  Write a list of blocks to a file.

def save_blocks (blocks, filename=None, encoding=None):
    f = outfile(filename, encoding)
    write_blocks(blocks, f)
    return close(f)

def write_blocks (blocks, f):
    first = True
    for block in blocks:
        write_block(block, f, first=first)
        first = False

def write_block (block, f, first=False):
    if not first:
        f.write('\n')
    for line in block:
        f.write(line)
        f.write('\n')


#--  Record Blocks  ------------------------------------------------------------

##  A record is a list of field values (strings).  A record block is a list of
#   records.  A record-block file is a list of record blocks.
#   In the file, fields are separated by tab and blocks are separated by empty lines.

def load_record_blocks (filename, encoding=None):
    return list(iter_record_blocks(filename, encoding))

##  Iterate over record blocks.

def iter_record_blocks (filename, encoding=None):
    return read_record_blocks(infile(filename, encoding))

def read_record_blocks (f):
    return list(_read_record_blocks(f))

def _read_record_blocks (f):
    for block in read_blocks(f):
        yield [tuple(line.split('\t')) for line in block]

##  Save a list of record blocks to a file.

def save_record_blocks (blocks, filename=None, encoding=None):
    write_record_blocks(blocks, outfile(filename, encoding))

def write_record_blocks (blocks, f):
    first = True
    for block in blocks:
        write_record_block(block, f, first)
        first = False

def write_record_block (block, f, first=False):
    write_block(['\t'.join(r) for r in block], f, first)


#--  Tokens  -------------------------------------------------------------------

##  Quote characters: single and double quote.
QuoteChars = '"\''

##  Default special characters: parentheses, brackets, braces.
DefaultSpecialChars = '()[]{}'


##  Represents the syntax of a little language.

class Syntax (object):

    ##  Constructor.

    def __init__ (self,
                  special=DefaultSpecialChars,
                  eol=False,
                  comments=True,
                  multi=None,
                  backslash=True,
                  stringtype='word',
                  mlstrings=False,
                  digits=False):

        ##  The special characters.  True means everything except alphanumerics
        #   and whitespace.
        self.special = None

        ##  Multi-character sequences that should be recognized as tokens.
        self.multi = None

        ##  The start characters for multi-character sequences.
        self.multi_start = None

        ##  Whether to include newline as a token.
        self.eol = eol

        ##  What character/string introduces comments.  Default: '#'.
        self.comments = None

        ##  Whether to interpret backslash as escape.
        self.backslash = backslash

        ##  What token type to use for a quoted string.  Default: 'word'.
        self.stringtype = stringtype

        ##  Whether to permit strings to extend over multiple lines.  Default: False.
        self.mlstrings = mlstrings

        ##  Whether to return digit sequences as tokens.
        self.digits = digits


        if comments is True:
            self.comments = '#'
        elif not comments:
            self.comments = ''

        if special is True:
            # everything is special except alphanum and underscore
            self.special = True
        else:
            self.special = special + QuoteChars + self.comments

        self.multi = multi
        if multi:
            self.multi = sorted(multi, key=len, reverse=True)
            self.multi_start = ''.join(set(s[0] for s in multi))
        else:
            self.multi_start = ''

    ##  Convert s to a string that will scan correctly.

    def scanable_string (self, s):
        if not isinstance(s, str):
            s = str(s)
        for c in s:
            if c.isspace() or c in self.special:
                s = repr(s)
                if s[0] == 'u': return s[1:]
                else: return s
        return s

    ##  String representation.

    def __repr__ (self):
        return '<Syntax %s %s>' % (self.special, self.eol)

##  A Syntax instance.
DefaultSyntax = Syntax()

##  Convert s to a scanable string, assuming the default syntax.

def scanable_string (s):
    return DefaultSyntax.scanable_string(s)


##  A token.

class Token (str):

    ##  True if type is 'any' or equals the token's type.
    def hastype (self, type):
        if type == 'any': return self.type != 'eof'
        else: return type == self.type

    ##  Raise an exception, indicating file, line, and offset.
    def error (self, msg='syntax error'):
        raise Exception('[%s line %d char %d] %s' % \
            (self.filename, self.line, self.offset, msg))

    ##  Print a warning, indicating file, line, and offset.
    def warn (self, msg):
        sys.stderr.write('WARNING: [%s line %d char %d] %s\n' % \
                             (self.filename, self.line, self.offset, msg))


##  Tokenize a file, returning a list.

def load_tokens (filename, **kwargs):
    return list(iter_tokens(filename, **kwargs))

##  Iterate over tokens of a string.

def tokenize (s, syntax=DefaultSyntax, encoding=None):
    return iter_tokens(StringIO(s), syntax, encoding)

##  Iterator over tokens.

class iter_tokens (object):

    ##  Constructor.

    def __init__ (self, filename, syntax=DefaultSyntax, encoding=None):

        ##  A Syntax instance.
        self.syntax = syntax

        ##  Stack, for (possibly nested) temporary syntax changes.
        self.stack = []

        ##  Filename string; set even for streams not associated with files.
        self.filename = filename_string(filename)

        ##  The lines of the file.
        self.lines = load_lines(filename, encoding)  # no newlines in lines

        ##  The current line.
        self.line = None

        ##  Current line number.
        self.linecount = 1

        ##  Current character offset on the line.
        self.offset = 0

        ##  Previous linecount.
        self.old_linecount = 1

        ##  Previous offset.
        self.old_offset = 0

        self.__token = None

        if self.lines: self.line = self.lines[0]
        else: self.line = ''

    ##  Returns self.

    def __iter__ (self):
        return self

    ##  Whether we are at EOF.

    def __bool__ (self):
        return self.token().type != 'eof'

    def __readline (self):
        if self.linecount < len(self.lines):
            self.line = self.lines[self.linecount]
            self.linecount += 1
        elif self.linecount == len(self.lines):
            self.line = ''
            self.linecount += 1
        else:
            raise Exception('Readline after EOF')
        self.offset = 0

    def __at_eof (self):
        return self.linecount > len(self.lines)

    def __empty_line (self):
        for c in self.line:
            if c in self.syntax.comments: return True
            elif not c.isspace(): return False
        return True

    def __is_special (self, c):
        if self.syntax.special is True:
            return not (c.isalnum() or c == '_')
        else:
            return c in self.syntax.special

    ##  Advance, if necessary, then return the current token.

    def token (self):
        if self.__token is None: self.__advance()
        return self.__token

    def __skip_eol (self):
        if self.offset >= len(self.line):
            if self.syntax.eol and not self.__empty_line():
                self.__set_token('\n', self.offset, string='\n')
                self.__readline()
            else:
                while self.offset >= len(self.line):
                    if self.__at_eof():
                        self.__set_token('eof', self.offset)
                        break
                    self.__readline()

    def __advance (self):
        self.old_linecount = self.linecount
        self.old_offset = self.offset
        self.__token = None
        try:
            while self.__token is None:
                self.__skip_eol()
                if self.__token is not None: return
                c = self.line[self.offset]

                if c in self.syntax.multi_start and self.__scan_multi(): pass
                elif c in self.syntax.comments: self.offset = len(self.line)
                elif c == "'" or c == '"': self.__scan_quoted()
                elif c.isspace(): self.offset += 1
                elif self.__is_special(c): self.__scan_special()
                elif self.syntax.digits and self.__is_digit(c): self.__scan_digit()
                else: self.__scan_word()

        except StopIteration:
            raise Exception('[%s line %d offset %d] Unexpected eof' % \
                (self.filename, self.linecount, self.offset))

    def __retreat (self):
        self.__token = None
        self.linecount = self.old_linecount
        self.offset = self.old_offset
        if self.linecount > 0:
            self.line = self.lines[self.linecount-1]
        else:
            self.line = None

    def __set_token (self, type, start, line=None, string=None, quotes=None):
        if line is None:
            line = self.linecount
        if string is None:
            string = self.line[start:self.offset]
        self.__token = Token(string)
        self.__token.type = type
        self.__token.filename = self.filename
        self.__token.line = line
        self.__token.offset = start
        self.__token.quotes = quotes

    def __is_digit (self, c):
        if c.isdigit(): return True
        i = self.offset + 1
        return c == '-' and i < len(self.line) and self.line[i].isdigit()

    def __scan_digit (self):
        start = self.offset
        if self.line[self.offset] == '-': self.offset += 1
        while self.offset < len(self.line) and self.line[self.offset].isdigit():
            self.offset += 1
        self.__set_token('digit', start)

    def __scan_word (self):
        start = self.offset
        while self.offset < len(self.line):
            c = self.line[self.offset]
            if c.isspace() or self.__is_special(c): break
            self.offset += 1
        self.__set_token('word', start)

    def __error (self, start, msg):
        raise Exception('[%s line %d char %d] %s' % \
            (self.filename, self.linecount, start, msg))

    def __scan_quoted (self):
        delim = self.line[self.offset]
        self.offset += 1
        start = self.offset
        restart = self.offset
        frags = []
        while True:
            while self.offset >= len(self.line):
                if self.syntax.mlstrings:
                    if restart < len(self.line):
                        frags.append(self.line[restart:])
                    frags.append('\n')
                    self.__readline()
                    restart = self.offset
                    if self.__at_eof():
                        self.__error(start, 'Unterminated string at EOF')
                else:
                    self.__error(start, 'End of line in string')
            c = self.line[self.offset]
            if c == delim:
                frags.append(self.line[restart:self.offset])
                self.offset += 1
                break
            elif c == '\\' and self.syntax.backslash:
                frags.append(self.line[restart:self.offset])
                frags.append(self.__scan_escape_sequence())
                restart = self.offset
            else:
                self.offset += 1
        self.__set_token(self.syntax.stringtype, start, self.linecount, ''.join(frags), delim)

    def __scan_escape_sequence (self):
        # self.line[self.offset] is backslash
        self.offset += 1
        if self.offset >= len(self.line): self.__error('Bad escape sequence')
        c = self.line[self.offset]
        self.offset += 1
        if c == '\\' or c == '"' or c == "'": return c
        elif c == 'a': return '\a'
        elif c == 'b': return '\b'
        elif c == 'f': return '\f'
        elif c == 'n': return '\n'
        elif c == 'r': return '\r'
        elif c == 't': return '\t'
        elif c == 'u':
            i = self.offset
            self.offset += 4
            if self.offset > len(self.line): self.__error('Bad escape sequence')
            return chr(int(self.line[i:self.offset], 16))
        elif c == 'U':
            self.__error('\\U escapes not implemented')
        elif c == 'v': return '\v'
        elif '0' <= c <= '7':
            i = self.offset
            self.offset += 1
            n = 1
            while n < 3 and self.offset < len(self.line) and \
                    '0' <= self.line[self.offset] <= '7':
                self.offset += 1
                n += 1
            return chr(int(self.line[i:self.offset], 8))
        elif c == 'x':
            i = self.offset
            self.offset += 1
            if self.offset < len(self.line) and \
                    ('0' <= self.line[self.offset] <= '9' or \
                     'a' <= self.line[self.offset] <= 'f' or \
                     'A' <= self.line[self.offset] <= 'F'):
                self.offset += 1
            d = int(self.line[i:self.offset], 16)
            if d < 0x100: return chr(d)
            else: return chr(d)

    def __scan_special (self):
        start = self.offset
        self.offset += 1
        self.__set_token(self.line[start], start)

    def __looking_at (self, word):
        for i in range(len(word)):
            t = self.offset + i
            if t >= len(self.line): return False
            if self.line[t] != word[i]: return False
        return True

    def __scan_multi (self):
        for word in self.syntax.multi:
            if self.__looking_at(word):
                start = self.offset
                self.offset += len(word)
                self.__set_token(self.line[start:self.offset], start)
                return True

    ##  Whether the next token is something other than EOF.
    #   If type or string is provided, the value indicates whether the next
    #   token has the given type and/or string.

    def has_next (self, type=None, string=None):
        if string:
            if type: raise Exception("Provide only one argument")
            return self.token() == string
        elif type:
            return self.token().hastype(type)
        else:
            return self.token().type != 'eof'

    ##  Iterator method.

    def __next__ (self):
        token = self.token()
        if token.type == 'eof': raise StopIteration
        self.__token = None
        self.old_linecount = self.linecount
        self.old_offset = self.offset
        return token

    ##  If the next token matches the given type and/or string, return it
    #   and advance.  Otherwise, return None.

    def accept (self, type=None, string=None):
        token = self.token()
        if type and not token.hastype(type):
            return None
        if string and not (token == string):
            return None
        return next(self)

    ##  If the next token has the given type and/or string, return it
    #   and advance.  Otherwise, signal an error.  Returns None at EOF.

    def require (self, type=None, string=None):
        token = self.token()
        if type and not token.hastype(type):
            token.error('Expecting ' + repr(type))
        if string and not (token == string):
            token.error('Expecting ' + repr(string))
        if type == 'eof': return None
        else: return next(self)

    ##  Signal an error, indicating filename and line number.

    def error (self, msg=None):
        token = self.token()
        token.error(msg)

    ##  Print a warning, showing filename and line number.

    def warn (self, msg=None):
        token = self.token()
        token.warn(msg)

    ##  Push the current syntax on the stack and switch to the given syntax.

    def push_syntax (self, syntax):
        self.stack.append(self.syntax)
        self.syntax = syntax
        self.__retreat()

    ##  Restore the previous syntax from the stack.

    def pop_syntax (self):
        if not self.stack: raise Exception('Empty stack')
        self.syntax = self.stack.pop()
        self.__retreat()


#--  Indenter  -----------------------------------------------------------------

##  Indenter.
#   The main facility a formatter provides is automatic indentation.
#   There is a prevailing indentation level, and indentation spaces are
#   automatically inserted after each newline that is written to the
#   formatter.  The level of indentation is increased
#   using begin_indent() and decreased
#   using end_indent().  It is initially zero.
#
#   A formatter may be turned off, in which case writing commands are
#   accepted but generate no output.  The formatter is initially on.

class Indenter:

    ##  Constructor.

    def __init__ (self, filename=None, encoding=None):
        self.__out = outfile(filename, encoding)

        ##  Whether it is enabled.
        self.enabled = True

        self.__at_bol = True
        self.__indent_level = 0
        self.__indent_string = ""

        ##  How many spaces to indent by each time.
        self.indent_step = 3

    ##  Flush any pending output.
        
    def flush (self):
        if self.__out: self.__out.flush()

    ##  Close the file.

    def close (self):
        if self.__out:
            self.__out.close()
            self.__out = None

    ##  Turn the indenter on.

    def on (self):
        self.enabled = True

    ##  Turn the indenter off.

    def off (self):
        self.enabled = False

    ##  Write str(x) to the output, inserting indentation at the beginning
    #   of each line.

    def write (self, x):
        string = str(x)
        if self.enabled and string:
            lines = string.split("\n")
            self.__write(lines[0])
            for i in range(1, len(lines)):
                self.newline()
                self.__write(lines[i])

    def __write (self, string):
        if string:
            if self.__at_bol: self.__out.write(self.__indent_string)
            self.__out.write(string)
            self.__at_bol = False

    ##  Equivalent to write('\n').

    def newline (self):
        self.__out.write("\n")
        self.__at_bol = True

    ##  Increase the indentation level by one.

    def begin_indent (self):
        self.__indent_level += 1
        self.__indent_string = " " * (self.indent_step * self.__indent_level)

    ##  Decrease the indentation level by one.

    def end_indent (self):
        self.__indent_level -= 1
        if self.__indent_level < 0: raise Exception("No indentation to end")
        self.__indent_string = " " * (self.indent_step * self.__indent_level)

    ##  Set the indentation level to zero.

    def reset (self):
        self.__indent_level = 0

    ##  If the internal stream is a StringIO, fetch its value.

    def getvalue (self):
        return self.__out.getvalue()


#--  pprint  -------------------------------------------------------------------

##  A PPrinter instance.
pprint = None


##  Indentation for a PPrinter.

class PPrintIndent (object):

    ##  Constructor.

    def __init__ (self, pprinter, n):

        ##  The pprinter.
        self.pprinter = pprinter

        ##  The number of spaces indented by.
        self.n = n

    ##  Enter.

    def __enter__ (self):
        self.pprinter._indent += self.n

    ##  Exit.

    def __exit__ (self, t, v, tb):
        self.pprinter._indent -= self.n
        

##  A color for a PPrinter.

class PPrintColor (object):

    ##  Constructor.

    def __init__ (self, pprinter, color):

        ##  The pprinter.
        self.pprinter = pprinter

        ##  The color.
        self.color = color

        ##  Saves the previous color.
        self.prevcolor = None

    def _set_color (self, color):
        self.pprinter._color = color
        sys.stdout.write(fg[color])
        sys.stdout.flush()

    ##  Enter.

    def __enter__ (self):
        self.prevcolor = self.pprinter._color or 'default'
        self._set_color(self.color)

    ##  Exit.

    def __exit__ (self, t, v, tb):
        self._set_color(self.prevcolor)


##  A pretty-printer.

class PPrinter (object):

    ##  Constructor.

    def __init__ (self, file=None):
        self._color = None
        self._indent = 0
        self._atbol = True
        self._brflag = False
        self._file = file
    
    ##  String representation.

    def __repr__ (self):
        return '<PPrinter %s>' % repr(self._file)

    ##  The protected file.

    def file (self):
        if self._file is None: return sys.stdout
        else: return self._file

    ##  Returns an indentation.  Calling this in a with-statement causes the
    #   indentation to be active within the body.

    def indent (self, n=2):
        return PPrintIndent(self, n)

    ##  Start an indentation.  One should usually do "with pp.indent()" instead.

    def start_indent (self, n=2):
        self._indent += n
    
    ##  End an indentation.

    def end_indent (self, n=2):
        self._indent -= n
        if self._indent < 0: self._indent = 0
    
    ##  Freshline.

    def br (self):
        self._brflag = True

    ##  Returns a color.  Calling this in a with-statement causes the color
    #   to be used in the body.

    def color (self, c):
        return PPrintColor(self, c)

    ##  Like print().  Handles embedded newlines correctly.

    def __call__ (self, *args, end=None, color=None):
        if color is None:
            self._call1(args, end)
        else:
            with self.color(color):
                self._call1(args, end)
                
    def _call1 (self, args, end):
        if end is None and (len(args) == 0 or not hasattr(args[-1], '__pprint__')):
            end = '\n'
        first = True
        for arg in args:
            if first: first = False
            else: self.file().write(' ')
            self._printarg(arg)
        if end:
            self._printarg(end)

    def _printarg (self, arg):
        if hasattr(arg, '__pprint__'):
            arg.__pprint__()
        else:
            self.write(str(arg))

    ##  Write a string.

    def write (self, s):
        f = self.file()
        i = 0
        n = len(s)
        while i < n:
            j = s.find('\n', i)
            if j < 0: j = n
            # it is possible that s[0] == '\n'
            if i < j:
                if self._brflag and not self._atbol:
                    f.write('\n')
                    self._atbol = True
                if self._atbol:
                    f.write(' ' * self._indent)
                f.write(s[i:j].replace('\x1b', '\\x1b'))
                self._atbol = False
            # if j < n then s[j] == '\n'
            if j < n:
                f.write('\n')
                j += 1
                self._atbol = True
            i = j
    
    ##  Emit a newline, unless the last character printed was a newline.

    def freshline (self):
        if not self._atbol:
            self.write('\n')

    ##  Print and immediately flush.

    def now (self, *args, end=''):
        self.__call__(*args, end)
        self.file().flush()
    
    ##  Flush the output.

    def flush (self):
        self.file().flush()


pprint = PPrinter()


#--  tabular  ------------------------------------------------------------------

def __getcell (row, j, tostring):
    if len(row) > j: return tostring(row[j])
    else: return ''

def __compute_width (rows, j, tostring):
    return max(len(__getcell(r,j,tostring)) for r in rows)

def colalign (rows, header=None, indent='', tostring=str, hlines=True):
    tmp = []
    if header: tmp.append(header)
    tmp.extend(rows)
    rows = tmp
    ncols = max(len(r) for r in rows)
    widths = [__compute_width(rows,j,tostring) for j in range(ncols)]
    for i in range(len(rows)):
        line = indent
        for j in range(ncols):
            if j: line += ' '
            line += '%-*s' % (widths[j], __getcell(rows[i], j, tostring))
        rows[i] = line
    if hlines:
        linewidth = sum(widths) + ncols - 1
        hline = '-' * linewidth
        if header:
            rows[1:1] = [hline]
        rows[-1:-1] = [hline]
    return rows

##  Produces a string representation of a table with aligned columns,
#   suitable for printing.

def tabular (rows, header=None, indent='', tostring=str, hlines=True):
    return '\n'.join(colalign(rows, header, indent, tostring, hlines))


#--  wget  ---------------------------------------------------------------------

##  Simple interface to urlretrieve().

def wget (url):
    return urllib.request.urlretrieve(url)[0]


#--  OutputRedirected  ---------------------------------------------------------

##  Does the actual output redirection.
class _Redirect (object):

    ##  Constructor.
    def __init__ (self):
        ##  The output file.
        self.output = None

    ##  Call it.
    def __call__ (self, f=None, mode=None):
        if f is None:
            return Redirection(StringIO(), True, self)
        elif mode:
            return Redirection(_open(f, mode), True)
        elif isinstance(f, str):
            mode = 'w'
            return Redirection(_open(f, mode), True)            
        else:
            return Redirection(f)


##  Output redirection.  One should not generally instantiate this class
#   directly.  Rather, use the redirect() function.

class Redirection (object):

    ##  Constructor.

    def __init__ (self, file, toclose=False, caller=None):

        ##  The file redirected to.
        self.file = file

        ##  Whether it should be closed when done.
        self.toclose = toclose

        ##  If caller is set, then caller.output gets set to file.getvalue() on exit.
        self.caller = caller

        ##  To save the old value of stdout.
        self.stdout = None

        ##  To save the old value of stderr.
        self.stderr = None

    ##  Enter.  Stdout and stderr are set to my file.

    def __enter__ (self):
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self.file
        sys.stderr = self.file
        return self

    ##  Exit.  If there is a caller, set caller.output to file.getvalue().
    #   If toclose is True, close the file.  Restore stdout and stderr.

    def __exit__ (self, etype, evalue, traceback):
        if self.caller is not None: self.caller.output = self.file.getvalue()
        if self.toclose: self.file.close()
        sys.stdout = self.stdout
        sys.stderr = self.stderr

    ##  Write a string to the file.

    def write (self, s):
        self.file.write(s)

    ##  Flush the file.

    def flush (self):
        self.file.flush()


##  Returns a Redirection, for use in a with-statement.
#   Arguments: filename and mode.
#   Call it with no arguments to get redirection to a string.
#   For example:
#
#       with redirect() as f:
#           f.write(...)
#           out = f.output
#

redirect = _Redirect()


#--  Location  -----------------------------------------------------------------

##  Affects some Location methods.
DryRun = False


##  Generalizes over local files and remote files represented by URL.

class Location (object):

    ##  Constructor.  The filename may contain a * wildcard.  It will be
    #   expanded out as the filename of an existing file.

    def __init__ (self, fn=None, user=None, host=None, pathname=None):

        ##  May include ~ and *
        self.orig_pathname = pathname

        if fn is None:

            ##  The specified user name, None for a local file.
            self.user = user

            ##  The specified host, None for a local file.
            self.host = host

            ##  The pathname, converted to absolute form.
            self.pathname = pathname

        else:
            if user or host or pathname:
                raise Exception('Specify either fn or user-host-pathname')

            i = fn.find('@')
            j = fn.find(':')
    
            # No colon: local file
            if j < 0:
                if i >= 0:
                    raise Exception('At-sign but no colon: ' + fn)
                self.user = None
                self.host = None
                self.pathname = fn
    
            # Remote file
            else:
    
                # User not specified: use current user
                if i < 0:
                    self.user = getpass.getuser()
                    self.host = fn[:j]
                else:
                    self.user = fn[:i]
                    self.host = fn[i+1:j]
                self.pathname = fn[j+1:]

        if self.host is None:
            if self.pathname.startswith('~'):
                fn = os.path.expanduser(self.pathname)
            else:
                fn = os.path.abspath(self.pathname)
            self.pathname = instantiate_pathname(fn)

    ##  String representation.

    def __repr__ (self):
        if self.user is None: u = ''
        elif isinstance(self.user, str): u = self.user + '@'
        else: u = '<BAD USER>@'
        if self.host is None: h = ''
        elif isinstance(self.host, str): h = self.host + ':'
        else: h = '<BAD HOST>:'
        if self.pathname is None: p = '<NO PATHNAME>'
        elif isinstance(self.pathname, str): p = self.pathname
        else: p = '<BAD PATHNAME>'
        return u + h + p

    ##  Join a component to create a new Location.

    def join (self, cpt):
        return Location(user=self.user, host=self.host, pathname=os.path.join(self.pathname, cpt))

    ##  Slash is a synonym for join().

    def __truediv__ (self, cpt):
        return self.join(cpt)

    ##  Creates a new Location by concatenating the given string to the pathname.
    #   Unlike join(), does not insert a slash.

    def __add__ (self, suffix):
        return Location(user=self.user, host=self.host, pathname=self.pathname+suffix)

    ##  Whether this Location is for a remote file.  Boolean not of is_local().

    def is_remote (self):
        return bool(self.host)

    ##  Whether this Location is for a local file.

    def is_local (self):
        return not self.host

    ##  Raise an exception unless this Location is a local file.

    def require_local (self):
        if self.is_remote(): raise Exception('Not a local file: %s' % self)

    ##  Signals an error if not local.
    #   Return is not guaranteed to be an *existing* file.

    def to_filename (self):
        self.require_local()
        return self.pathname

    ##  Open the file.

    def open (self, mode='r', encoding=None, makedirs=False):
        fn = self.to_filename()
        if makedirs:
            if not mode.startswith('w'):
                raise Exception('Makedirs=True but mode is not w')
            dir = os.path.dirname(fn)
            if not os.path.exists(dir):
                os.makedirs(dir)
        return _open(fn, mode, encoding=encoding)

    ##  Read the file.  Set encoding to 'bytes' to read a binary file.

    def read (self, encoding=None):
        if encoding == 'bytes':
            mode = 'rb'
            encoding = None
        else:
            mode = 'r'
        with self.open(mode, encoding) as f:
            return f.read()

    ##  Returns a TabularFile.

    def tabular (self, mode='r', encoding=None, separator='\t'):
        return TabularFile(self, mode, encoding, separator)

    ##  Calls load_dict() on the file.

    def dict (self):
        return load_dict(str(self))

    ##  Like dirname(), but returns a Location.

    def parent (self):
        if (not self.pathname) or self.pathname == '/':
            return None
        else:
            return Location(user=self.user, host=self.host,
                            pathname=os.path.dirname(self.pathname))

    ##  Like basename().  Returns a string.

    def name (self):
        return os.path.basename(self.pathname)

    ##  Returns (parent, name) pair.  The parent is a Location, the name
    #   is a string.

    def split (self):
        (dirname, name) = os.path.split(self.pathname)
        return (Location(user=self.user, host=self.host, pathname=dirname),
                name)

    ##  Whether the file exists.

    def exists (self):
        return os.path.exists(self.to_filename())

    ##  Whether the file is '/'.

    def is_root (self):
        return self.pathname == '/'

    ##  Mac-specific.  If the filename begins with '/Volumes', check whether
    #   the subdirectory within /Volumes exists.

    def is_mounted (self):
        if self.is_remote(): raise Exception('Not a local file')
        f0 = self
        f1 = f2 = None
        while not f0.is_root():
            f2 = f1
            f1 = f0
            f0 = f0.parent()
            if f0 is None: return True
        if f1.pathname == '/Volumes':
            return f2 is not None and f2.exists()
        else:
            return True

    ##  Whether the file is a symbolic link.

    def islink (self):
        os.path.islink(self.to_filename())

    ##  Whether the file is a directory.

    def isdir (self):
        return os.path.isdir(self.to_filename())

    ##  Return an iteration over file names in this directory.
    #   Raises an exception if this is a regular file.

    def listdir (self):
        dir = self.to_filename()
        if not os.path.exists(dir):
            return iter([])
        elif not os.path.isdir(dir):
            raise Exception('Not a directory: %s' % self)
        else:
            return iter(os.listdir(dir))

    ##  Like listdir(), but iterates over (name, loc) pairs.

    def items (self):
        for name in self.listdir():
            yield (name, self.join(name))

    ##  Returns the size of the file, in bytes.

    def size (self):
        return os.stat(self.to_filename()).st_size

    ##  Returns the modification time of the file, in seconds since the epoch.

    def modtime (self):
        return os.stat(self.to_filename()).st_mtime

    ##  Returns the current user's relation to the file: 'owner', 'group', or 'other'.

    def my_relation (self):
        return self._my_relation1(os.stat(self.to_filename()))

    def _my_relation1 (self, s):
        if s.st_uid == os.getuid(): return 'owner'
        elif s.st_gid in os.getgroups(): return 'group'
        else: return 'other'

    def _Xable (self, forwhom, u, g, o):
        s = os.stat(self.to_filename())
        m = s.st_mode
        if forwhom == 'me': forwhom = self._my_relation1(s)
        if forwhom == 'owner': return m & u
        elif forwhom == 'group': return m & g
        elif forwhom == 'other': return m & o
        else: raise Exception('Bad forwhom: %s' % forwhom)

    ##  Whether the file is readable.

    def readable (self, forwhom='me'):
        return self._Xable(forwhom, stat.S_IRUSR, stat.S_IRGRP, stat.S_IROTH)

    ##  Whether it is writable.

    def writable (self, forwhom='me'):
        return self._Xable(forwhom, stat.S_IWUSR, stat.S_IWGRP, stat.S_IWOTH)

    ##  Whether it is executable.

    def executable (self, forwhom='me'):
        return self._Xable(forwhom, stat.S_IXUSR, stat.S_IXGRP, stat.S_IXOTH)

    ##  The permissions value: an int representing a 3-digit octal number.

    def permissions (self):
        return stat.filemode(os.stat(self.to_filename()).st_mode)

    ##  The MD5 hash of the file.

    def md5 (self, silent=False):
        if not silent: print('Computing md5 hash for', self, '...', end='')
        sys.stdout.flush()
        s = subprocess.getoutput('openssl md5 < ' + self.to_filename())
        if s.startswith('(stdin)= '):
            s = s[9:]
        if not silent: print(' ok')
        return s

    ##  Whether the file has dir as an ancestor.

    def is_under (self, dir):
        return (self.host == dir.host and
                self.user == dir.user and
                self.pathname.startswith(dir.pathname))

    ##  Execute the file using os.system().

    def __call__ (self):
        return os.system(self.to_filename()) == 0


    #--  Modifications  --------------------

    def _actionmask (self, s, action, forwhom):
        w = 0
        if isinstance(forwhom, str): forwhom = [forwhom]
        for name in forwhom:
            if name == 'me': name = self._my_relation1(s)
            if name == 'all': w |= (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
            elif name == 'owner': w |= stat.S_IRWXU
            elif name == 'group': w |= stat.S_IRWXG
            elif name == 'other': w |= stat.S_IRWXO
            else: raise Exception('Bad name: %s' % name)
        a = 0
        if 'r' in action:
            a |= (stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
        if 'w' in action:
            a |= (stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
        if 'x' in action:
            a |= (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
        return w & a

    ##  Change permissions to allow the given action: 'r', 'w', 'x'.
    #   @arg forwhom - may be 'me', 'owner', 'group', 'other', or 'all'.

    def permit (self, action, forwhom='me', recurse=False, silent=False):
        return self._chmod('permit', action, forwhom, recurse, silent)

    ##  Change permissions to disallow the given action.

    def deny (self, action, forwhom='all', recurse=False, silent=False):
        return self._chmod('deny', action, forwhom, recurse, silent)

    def _chmod (self, change, action, forwhom, recurse, silent):
        if not silent: pprint(change.title(), self, action, forwhom)
        fn = self.to_filename()
        s = os.stat(fn, follow_symlinks=False)
        mask = self._actionmask(s, action, forwhom)

        # specifying 'me' on a recursive call chooses owner/group/other
        # based on ownership of top file; may be wrong if lower files have
        # different owners

        if recurse and self.isdir():
            for (dir, subdirs, files) in os.walk(fn):
                self._chmod1(change, dir, mask)
                for name in files:
                    fn = os.path.join(dir, name)
                    self._chmod1(change, fn, mask)
        else:
            self._chmod1(change, fn, mask, s)
        
    def _chmod1 (self, change, fn, mask, s=None):
        if s is None: s = os.stat(fn, follow_symlinks=False)
        if change == 'permit': mode = (s.st_mode | mask)
        elif change == 'deny': mode = (s.st_mode & ~mask)
        else: raise Exception('Bad change: %s' % change)
        os.chmod(fn, mode, follow_symlinks=False)

    ##  Create any missing directories above this file.

    def assure_parent (self):
        dir = os.path.dirname(self.to_filename())
        if not os.path.exists(dir):
            os.makedirs(dir)

    ##  Create this directory, including any missing directories above it.
    #   Prints out a message.  Can get just the message and not the action
    #   by setting DryRun = True.

    def make_directory (self, silent=False):
        global DryRun
        if not self.exists():
            if not silent: pprint('Make Directory', self)
            fn = self.to_filename()
            if DryRun: return
            os.makedirs(fn)

    ##  Copy this file to the target.  Can be disabled by DryRun.

    def copy_to (self, tgt, silent=False):
        global DryRun
        if not silent: pprint('Copy File', self, tgt)
        if self.is_remote() and tgt.is_remote():
            raise Exception('Not implemented: copying remote to remote')
        if DryRun: return
        if self.is_remote():
            server = _intern_server(self.host, self.user)
            server.get(self.pathname, local=tgt.to_filename())
        elif tgt.is_remote():
            server = _intern_server(tgt.host, tgt.user)
            server.put(self.to_filename(), remote=tgt.pathname)
        else:
            shutil.copyfile(self.to_filename(), tgt.to_filename())

    ##  Copy the source file to this one.  Can be disabled by DryRun.

    def copy_from (self, src, silent=False):
        src.copy_to(self, silent)
    
    ##  Rename this file.  Can be disabled by DryRun.

    def move_to (self, tgt, silent=False):
        global DryRun
        if not silent: pprint('Move File', self, tgt)
        if self.is_remote() or tgt.is_remote():
            raise Exception('Only local moves are possible')
        if DryRun: return
        shutil.move(self.to_filename(), tgt.to_filename())

    ##  Delete this file.  Can be disabled by DryRun.

    def delete_file (self, silent=False):
        if not silent: pprint('Delete File', self)
        if DryRun: return
        if self.is_remote():
            server = _intern_server(self.host, self.user)
            server('rm ' + self.pathname)
        else:
            os.unlink(self.to_filename())
    
    ##  Delete this directory, if it is empty, error otherwise.  Can be disabled
    #   by DryRun.

    def delete_empty_directory (self, silent=False):
        if not silent: pprint('Delete Empty Directory', self)
        if DryRun: return
        if self.is_remote():
            server = _intern_server(self.host, self.user)
            server('rmdir ' + self.pathname)
        else:
            os.unlink(self.to_filename())
    
    ##  Delete the directory hierarchy rooted here.
    #   Use with caution!
    
    def delete_hierarchy (self, silent=False):
        if self.is_remote():
            raise Exception('Not implemented: remote delete hierarchy')
        if not silent: pprint('Delete Hierarchy', self)
        top = self.to_filename()
    
        if DryRun: return
    
        if os.path.islink(top):
            os.unlink(top)
    
        elif os.path.isdir(top):
            for (dir, subdirs, files) in os.walk(top, topdown=False):
                for name in files:
                    fn = os.path.join(dir, name)
                    os.unlink(fn)
                # symbolic links to dirs show up here
                for name in subdirs:
                    fn = os.path.join(dir, name)
                    if os.path.islink(fn):
                        os.unlink(fn)
                os.rmdir(dir)

        else:
            os.unlink(top)

    ##  Same as delete_hierarchy(), but no error if the directory has already
    #   been deleted.  Also, refuses to delete the root directory.

    def clear (self, silent=False):
        if self.is_local() and self.exists():
            if not self.pathname or self.pathname == '/':
                raise Exception('Attempt to delete root')
            if self.isdir():
                self.delete_hierarchy(silent)
            else:
                self.delete_file(silent)

    ##  Same as clear(), but will even change the file permissions to override
    #   the lack of write permission.  Use with extreme caution.

    def nuke (self, silent=False):
        self.permit('w', 'me', recurse=True, silent=silent)
        self.parent().permit('w', 'me', silent=silent)
        self.clear(silent=silent)


##  Expand out a wildcard (*) by looking for an existing file matching the pattern.

def instantiate_pathname (orig):

    # if the * is in a directory name, we want the unique directory, even
    #     if the full pathname does not exist
    # from the end, in case there are multiple wildcards
        
    i = orig.rfind('*')
    if i < 0:
        return orig
    else:
        j = orig.find('/', i+1)
        if j > 0:
            suffix = orig[j:]
            fn = orig[:j]
        else:
            suffix = ''
            fn = orig
        alts = glob(fn)
        if len(alts) != 1:
            return orig
        return alts[0] + suffix

##  Returns the first pathname component.

def first_component (fn):
    i = fn.find('/')
    if i < 0: return fn
    else: return fn[:i]

def _make_writable (self, fn):
    if not os.path.islink(fn):
        mode = os.lstat(fn).st_mode
        os.chmod(fn, mode | 0o200)

            
##  Shorthand for Location.
def L (s): return Location(s)

# ##  The Seal destination directory, as a Location.
# Dest = L(__dest__)
# 
# ##  The Seal python directory, as a Location.
# Seal = L(__seal__)
# 
# ##  The Seal bin directory, a Location.
# Bin = Dest/'bin'
# 
# ##  The Seal examples directory, a Location.
# Examples = L(ex)
# 
# ##  The Seal data directory, a Location.
# Data = L(data)

##  The /tmp directory, as a Location.
Tmp = L('/tmp')


#--  TabularFile  --------------------------------------------------------------

##  A tabular file.

class TabularFile (object):

    ##  Constructor.
    #   Separator of None means any whitespace on read, single space on write.

    def __init__ (self, loc, mode='r', encoding=None, separator='\t'):

        ##  Location.
        self.location = loc

        ##  Mode.
        self.mode = mode

        ##  Character encoding.
        self.encoding = encoding

        ##  Field separator.
        self.separator = separator

        ##  Stream.
        self.file = None

    ##  Enter.

    def __enter__ (self):
        self.file = self.location.open(self.mode)
        return self

    ##  Exit.

    def __exit__ (self, t, v, tr):
        self.file.close()
        self.file = None

    ##  Returns self.

    def __iter__ (self):
        return self

    ##  Iterator method.

    def __next__ (self):
        return next(self.file).rstrip('\r\n').split(self.separator)

    ##  Like print().

    def write (self, *args):
        sep = self.separator or ' '
        last = len(args) - 1
        for (i, arg) in enumerate(args):
            self.file.write(str(arg))
            if i == last: self.file.write('\n')
            else: self.file.write(sep)


#--  Open  ---------------------------------------------------------------------

##  Wraps around stdin or stdout.

class PseudoFile (object):

    ##  Constructor.

    def __init__ (self, file):
        ##  File.
        self.file = file

    ##  Write to the file.

    def write (self, s):
        self.file.write(s)

    ##  Returns self.

    def __iter__ (self):
        return self

    ##  Iterator method.

    def __next__ (self):
        return self.file.__next__()

    ##  Enter.

    def __enter__ (self):
        return self

    ##  Exit.

    def __exit__ (self, t, v, tr):
        pass


##  Accepts '-' as filename, representing stdin/stdout.

def open (fn, mode='r', **kwargs):
    if fn == '-':
        if mode.startswith('r'): return PseudoFile(sys.stdin)
        else: return PseudoFile(sys.stdout)
    else:
        return _open(fn, mode, **kwargs)


#--  BackingSave  --------------------------------------------------------------

##  Handles locking (only).

class BackingAdmin (object):

    ##  Constructor.

    def __init__ (self, fn):

        ##  The filename.
        self.filename = fn

        ##  The lock, when locked.
        self.lock = None

    ##  Lock the file.

    def __enter__ (self):
        self.lock = open(fn + '.lock', 'w')
        flock(self.lock, LOCK_EX)
        return self

    ##  Release the lock.

    def __exit__ (self, type, value, tb):
        self.lock.close()
    
    ##  Replace the file with the backup version.

    def undo (self):
        bakfn = self.filename + '.bak'
        if not os.path.exists(bakfn):
            raise Exception('No undo information available: %s' % fn)
        if os.path.exists(self.filename):
            os.replace(self.filename, self.filename + '.redo')
        os.replace(bakfn, self.filename)

    ##  Restore the old version after an undo.

    def redo (self):
        redofn = self.filename + '.redo'
        if not os.path.exists(redofn):
            raise Exception('No redo information available')
        bakfn = self.filename + '.bak'
        if os.path.exists(self.filename):
            if os.path.exists(bakfn):
                os.unlink(bakfn)
            os.replace(self.filename, bakfn)
        os.replace(redofn, self.filename)


##  Handles backup and locking.

class BackingSave (object):

    ##  Constructor.

    def __init__ (self, fn, binary=False, makedirs=False):

        ##  Filename.
        self.filename = fn

        ##  The file.
        self.file = None

        ##  Whether to create missing directories.
        self.makedirs = makedirs

        ##  Whether it is a binary file.
        self.binary = binary

        ##  Write to a temp file, only touch the protected file if all writes
        #   complete successfully.
        self.tmp = fn + '.tmp'

        ##  The lock, while locked.
        self.lock = None

    ##  Enter.  Create any missing directories.  Lock the file.
    #   Open the temp file.  Return the temp file.

    def __enter__ (self):
        dir = os.path.dirname(self.filename)
        # if filename contains no slash, dir is ''
        if dir and not os.path.exists(dir):
            if self.makedirs:
                os.makedirs(dir)
            else:
                raise Exception('Directory does not exist: %s' % dir)
        self.lock = open(self.filename + '.lock', 'w')
        flock(self.lock, LOCK_EX)
        if self.binary: mode = 'wb'
        else: mode = 'w'
        self.file = open(self.tmp, mode)
        return self.file

    ##  Close the temp file.  If no error occurred, move the current
    #   file to backup and move the temp file into its place.
    #   If an error occurred, just delete the temp file.
    #   Release the lock.

    def __exit__ (self, type, value, traceback):
        self.file.close()
        if type is None:
            bak = self.filename + '.bak'
            if os.path.exists(bak):
                os.unlink(bak)
            if os.path.exists(self.filename):
                shutil.move(self.filename, bak)
            shutil.move(self.tmp, self.filename)
        else:
            if os.path.exists(self.tmp):
                os.unlink(self.tmp)
        self.lock.close()


#--  Indented lines  -----------------------------------------------------------

def _split_indent (line):
    i = 0
    while i < len(line) and line[i] == ' ':
        i += 1
    return (i, line[i:] if i > 0 else line)

def indented_lines (f):
    for line in f:
        line = line.strip('\r\n')
        yield _split_indent(line)


#--  KV format  ----------------------------------------------------------------
#
#  Values:
#      |    String, to end of line.
#      NUM  float or int; begins with +, -, or digit.
#      :    Object or list.  The type is determined by the first record.
#      {}   Empty object.
#      []   Empty list.
#      
#  An object consists of key-value pairs.
#  A list consists of values.
#
#  At beginning of file, or after a record whose value is ':', we are in "first-child" mode.
#  If the record begins with a key, the parent is an object.
#  If the file begins with a value, the parent is a list.
#  Otherwise, error.
#
#  Parent is the object that we are working on.
#  Parent_indent is parent's indentation.
#  Current_indent is the indentation for the current parent's *children*.
#  For the first child, current_indent is None.
#
#  If the parent is an object, the current record must be key-value.
#  If the parent is a list, the current record must be a value.
#
#  When the value is in ':', create a new parent.
#  Push the old parent and its indentation on the stack.
#

class KVISyntaxError (Exception): pass

def _kvexception (loc, msg):
    (fn, lno) = loc
    return KVISyntaxError(f'[{fn}:{lno}] {msg}')

def _kverror (loc, msg):
    raise _kvexception(loc, msg)
    
def _split_key_value (line):
    if not line[0].isalpha():
        return (None, line)
    i = 1
    while not (i >= len(line) or line[i].isspace()):
        i += 1
    if i == len(line):
        return (line, '')
    key = line[:i]
    while i < len(line) and line[i].isspace():
        i += 1
    return (key, line[i:])

def _parse_value (loc, value, parentpath):
    if not value:
        _kverror(loc, 'Empty value')
    elif value[0].isspace():
        _kverror(loc, 'Indentation must consist of spaces only')
    elif value[0] == '|':
        return value[1:]
    elif value == ':T':
        return True
    elif value == ':F':
        return False
    elif value == '-':
        return None
    elif value in ('{}', '{}:'):
        return {}
    elif value in ('[]', '[]:'):
        return []
    elif value == ':':
        return None
    elif value[0] in '/.~':
        if not (value.startswith('/') or
                value.startswith('./') or
                value.startswith('../') or
                value.startswith('~/') or
                value in ('.', '..', '~')):
            _kverror(loc, 'Illegal pathname')
        return value
    else:
        try:
            if '.' in value:
                return float(value)
            else:
                return int(value)
        except:
            raise _kvexception(loc, 'Illegal numeric value') from None

def read_kvi (f, filename=None):
    path = Path(filename).parent if filename else Path('.')
    roots = [None]
    stack = []
    parent_loc = (roots, 0)
    parent = None
    parent_indent = -1
    current_indent = None

    for (lno, (indent, line)) in enumerate(indented_lines(f), 1):
        loc = (filename, lno)
        if not line or line.startswith('#'):
            continue

        (key, value) = _split_key_value(line)

        # first child?
        if parent is None:
            parent = [] if key is None else {}
            (gp,k) = parent_loc
            gp[k] = parent

        if current_indent is None:
            if indent <= parent_indent:
                _kverror(loc, 'Illegal indentation: <= parent_indent')
            current_indent = indent
            
        # end of parent
        while indent < current_indent:
            if not stack:
                _kverror(loc, 'Illegal indentation: empty stack')
            current_indent = parent_indent
            (parent, parent_indent) = stack.pop()

        if indent != current_indent:
            _kverror(loc, 'Illegal indentation: > current_indent')

        # process record
        if key is None:
            if not isinstance(parent, list):
                _kverror(loc, 'Expecting a key-value pair')
            key = len(parent)
            parent.append(None)
        else:
            if not isinstance(parent, dict):
                _kverror(loc, 'Expecting a simple value')
            if key in parent:
                _kverror(loc, f'Key {key} is not unique')

        if not value:
            _kverror(loc, 'Missing value')

        parent[key] = parsed_value = _parse_value(loc, value, path)

        # push
        if value in (':', '[]:', '{}:'):
            parent_loc = (parent, key)
            stack.append((parent, parent_indent))
            parent = parsed_value
            parent_indent = current_indent
            current_indent = None

    return roots[0]

def load_kvi (fn, **kwargs):
    with open(fn, **kwargs) as f:
        return read_kvi(f, filename=fn)

def isdictlike (x):
    return hasattr(x, 'items')

def islistlike (x):
    return hasattr(x, '__iter__') and not isdictlike(x)

def write_kvi (x, f):
    if isdictlike(x):
        _write_dict_contents(x, f, 0)
    elif islistlike(x):
        _write_list_contents(x, f, 0)
    else:
        raise Exception('Contents of KVI must be a dict or list')

def _write_dict_contents (x, f, indent):
    for (k,v) in x.items():
        f.write(' ' * indent)
        f.write(k)
        f.write(' ')
        _write_value(v, f, indent)

def _write_list_contents (x, f, indent):
    for v in x:
        f.write(' ' * indent)
        _write_value(v, f, indent)

def _write_value (x, f, indent):
    if isinstance(x, dict):
        if len(x) == 0:
            f.write('{}\n')
        else:
            f.write('{}:\n')
            _write_dict_contents(x, f, indent + 2)
    elif isinstance(x, (list, tuple)):
        if len(x) == 0:
            f.write('[]\n')
        else:
            f.write('[]:\n')
            _write_list_contents(x, f, indent + 2)
    elif isinstance(x, str):
        if '\n' in x:
            raise Exception('Value may not contain newline')
        f.write('|')
        f.write(x)
        f.write('\n')
    elif isinstance(x, bool):
        f.write(':T\n' if x else ':F\n')
    elif isinstance(x, (int, float)):
        f.write(str(x))
        f.write('\n')
    elif x == None:
        f.write('-\n')
    else:
        raise Exception('Illegal object type for KVI')

def save_kvi (x, fn, **kwargs):
    with open(fn, 'w', **kwargs) as f:
        write_kvi(x, f)


#--  File Formats  -------------------------------------------------------------

class FileFormat (object):

    def __init__ (self, name=None, read=None, write=None, encoding=None):
        self.name = name
        self.read = read
        self.write = write
        self.encoding = encoding

    def load (self, fn):
        if self.encoding is False:
            mode = 'rb'
            enc = None
        else:
            mode = 'r'
            enc = self.encoding
        with open(fn, mode, encoding=enc) as f:
            return self.read(f)
            
    def parse (self, s):
        with StringIO(s) as f:
            return self.read(f)

    def save (self, x, fn):
        if self.encoding is False:
            mode = 'wb'
            enc = None
        else:
            mode = 'w'
            enc = self.encoding
        with open(fn, mode, encoding=enc) as f:
            self.write(x, f)


def _read_lines (f):
    return [line.rstrip('\r\n') for line in f]

def _write_lines (lines, f):
    for line in lines:
        f.write(line)
        f.write('\n')

def _read_tabular (f):
    return [line.rstrip('\r\n').split('\t') for line in f]

def _write_tabular (recs, f):
    for rec in recs:
        f.write('\t'.join(rec))
        f.write('\n')

LineFormat = FileFormat('lines', _read_lines, _write_lines)
TabularFormat = FileFormat('tabular', _read_tabular, _write_tabular)
KVIFormat = FileFormat('kvi', read_kvi, write_kvi)
JsonFormat = FileFormat('json', json.load, json.dump)
BlockFormat = FileFormat('blocks', read_record_blocks, write_record_blocks)
